
#include "lg/interface.h"
#include "lg/scrservices.h"
#include "ScriptLib.h"
#include "GenericScale.h"

#include <cstring>

cScrScale::~cScrScale()
{
	delete[] m_pszScaleStimMsg;
}

MSGHANDLER cScrScale::GenScaleBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!m_iScaleStim && ParamExists(m_iObjId, "arcontrol"))
	{
		int iStim = ParamGetObject(m_iObjId, "arcontrol");
		if (iStim)
		{
			SService<IObjectSrv> pOS(g_pScriptManager);
			cScrStr sObjName;
			pOS->GetName(sObjName, iStim);
			m_pszScaleStimMsg = new char[::strlen(sObjName) + 9];
			::strcpy(m_pszScaleStimMsg, sObjName);
			::strcat(m_pszScaleStimMsg, "Stimulus");
			RegisterMessageHandler(m_pszScaleStimMsg, static_cast<MessageHandlerFunc>(&cScrScale::GenScaleStimulus));
			m_iScaleStim = iStim;
		}
		else
			DebugString("arcontrol stimulus not found");
	}
	return 1;
}

MSGHANDLER cScrScale::GenScaleEndScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_iScaleStim)
	{
		UnregisterMessageHandler(m_pszScaleStimMsg, static_cast<MessageHandlerFunc>(&cScrScale::GenScaleStimulus));
		delete[] m_pszScaleStimMsg;
		m_pszScaleStimMsg = NULL;
		m_iScaleStim = 0;
	}
	return 1;
}

MSGHANDLER cScrScale::GenScaleStimulus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	long iRet = OnScale(static_cast<sStimMsg*>(pMsg)->intensity, pMsg, pReply);
	return iRet;
}

long cScrScale::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	long iRet = 1;
	if (ParamExists(m_iObjId, "on"))
	{
		float fScale = ParamGetFloat(m_iObjId, "on");
		iRet = OnScale(fScale, pMsg, pReply);
	}
	if (iRet && pReply)
		MultiParmAssignInt(pReply, 0);
	return iRet;
}

long cScrScale::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	long iRet;
	float fScale;
	if (! ParamExists(m_iObjId, "off"))
	{
		if (! ParamExists(m_iObjId, "on"))
		{
			if (pReply)
				MultiParmAssignInt(pReply, 0);
			return 1;
		}
		fScale = - ParamGetFloat(m_iObjId, "on");
	}
	else
		fScale = ParamGetFloat(m_iObjId, "off");
	iRet = OnScale(fScale, pMsg, pReply);
	if (iRet && pReply)
		MultiParmAssignInt(pReply, 0);
	return iRet;
}

long cScrScale::OnControl(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	float fScale;
	long iRet;
	switch (pMsg->data.type)
	{
	case kMT_Int:
		fScale = pMsg->data.i;
		break;
	case kMT_Float:
		fScale = pMsg->data.f;
		break;
	case kMT_String:
		if (pMsg->data.psz)
		{
			char* pszEnd;
			fScale = strtod(pMsg->data.psz, &pszEnd);
			if (pszEnd && *pszEnd == '\0')
				break;
		}
	default:
		return 1;
	}
	iRet = OnScale(fScale, pMsg, pReply);
	if (iRet && pReply)
		MultiParmAssignInt(pReply, 0);
	return iRet;
}

