
#include "lg/interface.h"
#include "lg/scrservices.h"
#include "ScriptLib.h"
#include "GenericScript.h"


void cScrGeneric::InitTrapVars()
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "ScriptTiming"))
	{
		cMultiParm mpTiming;
		pPropSrv->Get(mpTiming, m_iObjId, "ScriptTiming", NULL);
		m_iTiming = mpTiming;
	}
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		m_iFlags = static_cast<int>(mpTrapFlags);
	}
}

bool cScrGeneric::IsLocked(sScrMsg* pMsg)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	return pLock->IsLocked(m_iObjId);
}

void cScrGeneric::SetLock(bool bLock)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	int iLock = GetOneLinkDest("Lock", m_iObjId);
	if (iLock)
	{
		if (pPropSrv->Possessed(m_iObjId, "Locked"))
			pPropSrv->Remove(m_iObjId, "Locked");
		pPropSrv->SetSimple(iLock, "Locked", static_cast<int>(bLock));
	}
	else
	{
		pPropSrv->SetSimple(m_iObjId, "Locked", static_cast<int>(bLock));
	}
}

MSGHANDLER cScrGeneric::GenericTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	const char* pszMsgName = static_cast<const char*>(pTimerMsg->name);
	m_iTiming = 0;
	m_iFlags = 0;
	if (!stricmp(pszMsgName, "TurnOn"))
	{
		return OnTurnOn(pMsg, pReply);
	}
	else if (!stricmp(pszMsgName, "TurnOff"))
	{
		return OnTurnOff(pMsg, pReply);
	}
	return 1;
}

MSGHANDLER cScrGeneric::GenericTurnOn(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	InitTrapVars();
	if (m_iFlags & kTrapFlagNoOn)
		return 1;
	if (IsLocked(pMsg))
	{
		return 0;
	}
	long iRet;
	const char* pszInverseMsg;
	if (m_iFlags & kTrapFlagInvert)
	{
		iRet = OnTurnOff(pMsg, pReply);
		pszInverseMsg = "TurnOn";
	}
	else
	{
		iRet = OnTurnOn(pMsg, pReply);
		pszInverseMsg = "TurnOff";
	}
	if (iRet == 0)
	{
		if (m_iTiming > 0)
			g_pScriptManager->SetTimedMessage2(m_iObjId, pszInverseMsg, m_iTiming, kSTM_OneShot, 0);
		if (m_iFlags & kTrapFlagOnce)
		{
			SetLock(true);
		}
	}
	return iRet;
}

MSGHANDLER cScrGeneric::GenericTurnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	InitTrapVars();
	if (m_iFlags & kTrapFlagNoOff)
		return 1;
	if (IsLocked(pMsg))
	{
		return 0;
	}
	long iRet;
	//const char* pszInverseMsg;
	if (m_iFlags & kTrapFlagInvert)
	{
		iRet = OnTurnOn(pMsg, pReply);
		//pszInverseMsg = "TurnOff";
	}
	else
	{
		iRet = OnTurnOff(pMsg, pReply);
		//pszInverseMsg = "TurnOn";
	}
	if (iRet == 0)
	{
		if (m_iFlags & kTrapFlagOnce)
		{
			SetLock(true);
		}
	}
	return iRet;
}

MSGHANDLER cScrGeneric::GenericScriptControl(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	InitTrapVars();
	if (IsLocked(pMsg))
	{
		return 0;
	}
	long iRet = OnControl(pMsg, pReply);
	return iRet;
}

MSGHANDLER cScrGeneric::DelayedInit(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	if (stricmp(static_cast<const char*>(pTimerMsg->name), "DelayedInit"))
		return 1;
	if (pTimerMsg->data.type == kMT_String && 
		!stricmp(pTimerMsg->data.psz, "FixupPlayerLinks"))
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		object oPlayer;
		pOS->Named(oPlayer, "Player");
		if (oPlayer != 0)
			::FixupPlayerLinks(m_iObjId, oPlayer);
		UnregisterMessageHandler("Timer", static_cast<MessageHandlerFunc>(&cScrGeneric::DelayedInit));
	}
	return 0;
}

int cScrGeneric::FixupPlayerLinks(void)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	object oPlayer;
	pOS->Named(oPlayer, "Player");
	if (oPlayer != 0)
	{
		::FixupPlayerLinks(m_iObjId, oPlayer);
		return 0;
	}
	else
	{
		if (RegisterMessageHandler("Timer", static_cast<MessageHandlerFunc>(&cScrGeneric::DelayedInit)))
			g_pScriptManager->SetTimedMessage2(m_iObjId, "DelayedInit", 1, kSTM_OneShot, "FixupPlayerLinks");
		return 1;
	}
}

