
#ifndef GENERICSCRIPT_H
#define GENERICSCRIPT_H

#include "Script.h"


/*
 * Generic Script
 *
 */
class cScrGeneric : public cScript
{
public:
	cScrGeneric(const char* pszName, int iHostObjId)
		: cScript(pszName, iHostObjId),
		  m_iTiming(0), m_iFlags(0)
	{ }

private:
	DECLARE_MSGHANDLER(DelayedInit);
	DECLARE_MSGHANDLER(GenericTimer);
	DECLARE_MSGHANDLER(GenericTurnOn);
	DECLARE_MSGHANDLER(GenericTurnOff);
	DECLARE_MSGHANDLER(GenericScriptControl);

protected:
	virtual void InitScriptData()
	{
		cScript::InitScriptData();
		REGISTER_MSGHANDLER("Timer",cScrGeneric::GenericTimer);
		REGISTER_MSGHANDLER("TurnOn",cScrGeneric::GenericTurnOn);
		REGISTER_MSGHANDLER("TurnOff",cScrGeneric::GenericTurnOff);
		REGISTER_MSGHANDLER("ScriptControl",cScrGeneric::GenericScriptControl);
	}

	// arguments are the same from ReceiveMessage
	// The message is not guaranteed to be the same as the function,
	// because of the invert flag, timers, and ScriptControl messages.
	virtual long OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply) = 0;
	virtual long OnTurnOff(sScrMsg*,sMultiParm*) = 0;
	virtual long OnControl(sScrMsg*,sMultiParm*) = 0;
	// Copy the Script\TrapControlFlags to m_iFlags
	virtual void InitTrapVars();
	// Test if the object is locked.
	virtual bool IsLocked(sScrMsg* pMsg);
	virtual void SetLock(bool bLock);

	// The default handler will automatically turn off the trap if ScriptTiming is set.
	// You may modify this during your function. If it is zero, no timer is set.
	// The timer will not be set if your function returns non-zero.
	int m_iTiming;
	// The trap control flags. The Once flag is check after your function, so you can 
	// clear it to disable that feature.
	unsigned long m_iFlags;

public:
	int FixupPlayerLinks(void);

};

#endif // GENERICTRAP_H
