
#ifndef SCRIPT_H
#define SCRIPT_H

#include <map>
#include <list>
#include "strless_nocase.h"

#include "lg/objstd.h"
#include "lg/script.h"

#undef DispatchMessage

#ifdef GENOSM
# define DECLARE_MODULE(osmName)			DECLARE_MODULE(osmName)
# define DECLARE_SCRIPT(clName,clBase)		DECLARE_SCRIPT(clName,clBase)
# define DECLARE_FACTORY(scrName,scrBase)	DECLARE_FACTORY(scrName,scrBase)
# define DECLARE_MSGHANDLER(fnName)			DECLARE_MSGHANDLER(fnName)
# define REGISTER_MSGHANDLER(msgName,fnHandler)	REGISTER_MSGHANDLER(msgName,fnHandler)
#else
# define DECLARE_MODULE(osmName)
# define DECLARE_SCRIPT(clName,clBase)  class clName : public clBase
# define DECLARE_FACTORY(scrName,scrBase)  static IScript* __cdecl Make##scrName (const char* pszName, int iHostObjId); \
	static void InitScriptDesc (sNewClassDesc * pDesc)

# define MSGHANDLER  long __cdecl
# define DECLARE_MSGHANDLER(fnName)  MSGHANDLER fnName(sScrMsg*, sMultiParm*, eScrTraceAction)
# define REGISTER_MSGHANDLER(msgName,fnHandler)  \
	RegisterMessageHandler(msgName, static_cast<MessageHandlerFunc>(&fnHandler))
#endif


struct sNewClassDesc : public sScrClassDesc
{
	ScriptFactoryProc pfnRealFactory;
};


class cScript : public IScript
{
protected:
	typedef long (__cdecl cScript::*MessageHandlerFunc)(sScrMsg*, sMultiParm*, eScrTraceAction);

private:
	typedef std::map<const char*, std::list<MessageHandlerFunc>, strless_nocase>  tDispatchMap;

public:
	// IUnknown
	STDMETHOD(QueryInterface)(REFIID,void**);
	STDMETHOD_(ULONG,AddRef)(void);
	STDMETHOD_(ULONG,Release)(void);
	// IScript
	STDMETHOD_(const char*,GetClassName)(void);
	STDMETHOD_(long,ReceiveMessage)(sScrMsg*,sMultiParm*,eScrTraceAction);

private:
	long int m_iRef;

public:
	virtual ~cScript();
	cScript(const char* pszName, int iHostObjId);

	STDMETHOD_(void*,GetThisPointer)()
	{
		return reinterpret_cast<void*>(this);
	}
	void DebugString(const char* pszMsg1, const char* pszMsg2 = "");

protected:
	char* m_szName;
	int m_iObjId;
	bool m_bSim;

	tDispatchMap m_mapHandlers;

	bool RegisterMessageHandler(const char* pszMsgName, MessageHandlerFunc pmfMsgHandler);
	bool UnregisterMessageHandler(const char* pszMsgName, MessageHandlerFunc pmfMsgHandler);
	long DispatchMessage(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace);
	virtual void InitScriptData();

private:

	MSGHANDLER ScriptPtrQuery(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace);

};

#endif // SCRIPT_H
