#include <cstring>

#include "lg/interface.h"
#include "lg/scrservices.h"
#include "lg/scrmanagers.h"
#include "lg/objects.h"
#include "lg/links.h"
#include "lg/properties.h"
#include "lg/propdefs.h"
#include "ScriptLib.h"
#include "VendorLib.h"

int VendorGetStore(int iObj)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	linkset lsLinks;
	pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), iObj, 0);
	for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
	{
		if (!stricmp(reinterpret_cast<const char*>(lsLinks.Data()), "Store"))
		{
			sLink sl = lsLinks.Get();
			return sl.dest;
		}
	}
	return iObj;
}

int VendorGetCart(int iObj)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	linkset lsLinks;
	pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), iObj, 0);
	for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
	{
		if (!stricmp(reinterpret_cast<const char*>(lsLinks.Data()), "Cart"))
		{
			sLink sl = lsLinks.Get();
			return sl.dest;
		}
	}
	return 0;
}

int VendorGetItem(int iObj)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	linkset lsLinks;
	pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), iObj, 0);
	for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
	{
		if (!stricmp(reinterpret_cast<const char*>(lsLinks.Data()), "Item"))
		{
			sLink sl = lsLinks.Get();
			return sl.dest;
		}
	}
	return 0;
}

static int _GetLoot(int iObj, sLoot* pRet)
{
	int iLootObj = 0;
	SInterface<IPropertyManager> pPM(g_pScriptManager);
	IGenericProperty* pProp = static_cast<IGenericProperty*>(pPM->GetPropertyNamed("Loot"));
	if (pProp)
	{
		if (pProp->IsRelevant(iObj))
		{
			sLoot *pLoot = NULL;
			pProp->Get(iObj, reinterpret_cast<void**>(&pLoot));
			*pRet = *pLoot;
			iLootObj = iObj;
		}
		else
		{
			SService<ILinkSrv> pLS(g_pScriptManager);
			SService<ILinkToolsSrv> pLTS(g_pScriptManager);
			linkset lsLinks;
			pLS->GetAll(lsLinks, pLTS->LinkKindNamed("Contains"), iObj, 0);
			for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
			{
				sLink sl = lsLinks.Get();
				if (pProp->IsRelevant(sl.dest))
				{
					sLoot *pLoot = NULL;
					pProp->Get(sl.dest, reinterpret_cast<void**>(&pLoot));
					*pRet = *pLoot;
					iLootObj = sl.dest;
					break;
				}
			}
		}
		pProp->Release();
	}
	return iLootObj;
}

static void _SetLoot(int iObj, sLoot* pLoot)
{
	SInterface<IPropertyManager> pPM(g_pScriptManager);
	IGenericProperty* pProp = static_cast<IGenericProperty*>(pPM->GetPropertyNamed("Loot"));
	if (pProp)
	{
		if (!pProp->IsRelevant(iObj))
			pProp->Create(iObj);
		pProp->Set(iObj, reinterpret_cast<void*>(&pLoot));
		pProp->Release();
	}
}

int VendorGetLoot(int iObj, int* iGold, int* iGems, int* iGoods)
{
	if (!iObj)
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		object obj;
		pOS->Named(obj, "Player");
		iObj = obj;
	}
	sLoot loot = { 0, 0, 0, 0 };
	_GetLoot(iObj, &loot);

	if (iGold)
		*iGold = loot.iGold;
	if (iGems)
		*iGems = loot.iGems;
	if (iGoods)
		*iGoods = loot.iArt;
	return loot.iGold + loot.iGems + loot.iArt;
}

void VendorAddLoot(int iObj, int iGold, int iGems, int iGoods)
{
	if (!iObj)
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		object obj;
		pOS->Named(obj, "Player");
		iObj = obj;
	}
	sLoot loot = { 0, 0, 0, 0 };
	/* Live dangerously
	if (iGold < 0 || iGems < 0 || iGoods < 0)
	{
		_GetLoot(iObj, &loot);
		if (iGold < -loot.iGold)
			iGold = -loot.iGold;
		if (iGems < -loot.iGems)
			iGems = -loot.iGems;
		if (iGoods < -loot.iArt)
			iGoods = -loot.iArt;
	}
	*/
	SInterface<IObjectSystem> pOS(g_pScriptManager);
	SInterface<IPropertyManager> pPM(g_pScriptManager);
	SInterface<IContainSys> pCS(g_pScriptManager);
	IGenericProperty* pProp = static_cast<IGenericProperty*>(pPM->GetPropertyNamed("Loot"));
	if (pProp)
	{
		if (pProp->IsRelevant(iObj))
		{
			sLoot *pLoot = NULL;
			pProp->Get(iObj, reinterpret_cast<void**>(&pLoot));
			loot.iGold = pLoot->iGold + iGold;
			loot.iGems = pLoot->iGems + iGems;
			loot.iArt = pLoot->iArt + iGoods;
			loot.iSpecial = pLoot->iSpecial;
			pProp->Set(iObj, reinterpret_cast<void*>(&loot));
		}
		else
		{
			int iPurse = pOS->BeginCreate(pOS->GetObjectNamed("Purse"), 1);
			if (iPurse)
			{
				pProp->Create(iPurse);
				loot.iGold = iGold;
				loot.iGems = iGems;
				loot.iArt = iGoods;
				loot.iSpecial = 0;
				pProp->Set(iPurse, reinterpret_cast<void*>(&loot));
				pOS->EndCreate(iPurse);
				pCS->Add(iObj, iPurse, 0, 1);
			}
		}
		pProp->Release();
	}
}

int VendorDebitLoot(int iObj, int iAmount, eVendorDebitType eType)
{
	if (!iObj)
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		object obj;
		pOS->Named(obj, "Player");
		iObj = obj;
	}
	sLoot loot = { 0, 0, 0, 0 };
	sLoot debit = { 0, 0, 0, 0 };
	int iPurse = _GetLoot(iObj, &loot);
	if (!iPurse)
		return 0;
	switch (eType)
	{
	case kDebitExact:
	case kDebitGold:
		if (loot.iGold < iAmount)
			return 0;
		debit.iGold = -iAmount;
		debit.iGems = 0;
		debit.iArt = 0;
		break;
	case kDebitGems:
		if (loot.iGems < iAmount)
			return 0;
		debit.iGems = -iAmount;
		debit.iGold = 0;
		debit.iArt = 0;
		break;
	case kDebitGoods:
		if (loot.iArt < iAmount)
			return 0;
		debit.iArt = -iAmount;
		debit.iGold = 0;
		debit.iGems = 0;
		break;
	case kDebitGoldFirst:
		if (iAmount > (loot.iGold + loot.iGems + loot.iArt))
			return 0;
		if (iAmount > loot.iGold)
		{
			if ((iAmount - loot.iGold) > loot.iGems)
			{
				debit.iGold = -loot.iGold;
				debit.iGems = -loot.iGems;
				debit.iArt = loot.iGold + loot.iGems - iAmount;
			}
			else
			{
				debit.iGold = -loot.iGold;
				debit.iGems = loot.iGold - iAmount;
			}
		}
		else
			debit.iGold = -iAmount;
		break;
	case kDebitEvenly:
		{
			if (iAmount > (loot.iGold + loot.iGems + loot.iArt))
				return 0;
			div_t d = div(iAmount, 3);
			debit.iArt = debit.iGems = debit.iGold = -d.quot;
			if (loot.iArt < d.quot)
			{
				d.rem += d.quot - loot.iArt;
				debit.iArt = -loot.iArt;
				loot.iArt = 0;
			}
			else
				loot.iArt -= d.quot;
			if (loot.iGems < d.quot)
			{
				d.rem += d.quot - loot.iGems;
				debit.iGems = -loot.iGems;
				loot.iGems = 0;
			}
			else
				loot.iGems -= d.quot;
			if (loot.iGold < d.quot)
			{
				d.rem += d.quot - loot.iGold;
				debit.iGold = -loot.iGold;
				loot.iGold = 0;
			}
			else
				loot.iGold -= d.quot;
			while (d.rem-- > 0)
			{
				if (loot.iGold < loot.iGems || loot.iGold < loot.iArt)
				{
					if (loot.iGems < loot.iArt)
					{
						--loot.iArt;
						--debit.iArt;
					}
					else
					{
						--loot.iGems;
						--debit.iGems;
					}
				}
				else
				{
					--loot.iGold;
					--debit.iGold;
				}
			}
			break;
		}
	}
	VendorAddLoot(iObj, debit.iGold, debit.iGems, debit.iArt);
	return 1;
}

int VendorDebitLoot(int iObj, int iGold, int iGems, int iGoods)
{
	if (!iObj)
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		object obj;
		pOS->Named(obj, "Player");
		iObj = obj;
	}
	sLoot loot = { 0, 0, 0, 0 };
	int iPurse = _GetLoot(iObj, &loot);
	if (!iPurse)
		return 0;
	if (loot.iGold < iGold || loot.iGems < iGems || loot.iArt < iGoods)
		return 0;
	VendorAddLoot(iObj, -iGold, -iGems, -iGoods);
	return 1;
}


