#if _DARKGAME != 2
#error tnhScript is written for Thief2. You can try, but I doubt it.
#else

#include <cmath>
#include <cstdlib>
#include <cstring>
#include <cctype>
#include <vector>
#include <set>

#include <lg/interface.h>
#include <lg/scrservices.h>
#include <lg/scrmanagers.h>
#include <lg/scrmsgs.h>
#include <lg/properties.h>
#include <lg/propdefs.h>
#include <lg/objects.h>
#include <lg/input.h>
#include <lg/quest.h>

#include <darkhook.h>
#include "ScriptLib.h"
#include "tnhScript.h"


/*******************
 * ScriptController
 */
long cScr_ScriptCtrl::OnControlString(const char* pszControl, sScrMsg* pMsg, sMultiParm* pReply)
{
	int iDestObj = 0;
	const char* pszDest = static_cast<const char*>(pMsg->data2);
	if (pszDest && *pszDest)
	{
		iDestObj = StrToObject(pszDest, pMsg->to, pMsg->from);
	}
	if (iDestObj)
	{
		g_pScriptManager->SendMessage2(*static_cast<cMultiParm*>(pReply), m_iObjId, iDestObj, 
				"ScriptControl", pszControl, cMultiParm::Undef, cMultiParm::Undef);
	}
	else
	{
		CDSend("ScriptControl", m_iObjId, pszControl);
		if (pReply)
			MultiParmAssignInt(pReply, 1);
	}
	return 0;
}


/*******************
 * CommandControl
 */
long cScr_CommandCtrl::OnControlString(const char* pszControl, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IDebugScrSrv> pDebugSrv(g_pScriptManager);
	char* pszCommand;
	for (pszCommand = strtok(const_cast<char*>(pszControl), ";"); pszCommand; pszCommand = strtok(NULL, ";"))
	{
		pDebugSrv->Command(pszCommand, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null);
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*******************
 * TrapFrobber
 */
long cScr_TrapFrobber::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	sFrobMsg pFrobMsg;
	//sMultiParm mpRet;
	pFrobMsg.from = m_iObjId;
	pFrobMsg.to = m_iObjId;
	pFrobMsg.time = pMsg->time;
	pFrobMsg.flags = 4; // And someday I might even learn what this means.
	pFrobMsg.SrcObjId = m_iObjId;
	pFrobMsg.DstObjId = 0;
	pFrobMsg.Frobber = ParamGetBool(m_iObjId, "player_frob") ? 
			StrToObject("Player") : 0;
	pFrobMsg.DstLoc = kFrobLocNone;
	pFrobMsg.Sec = 0.0;
	pFrobMsg.Abort = 0;
	if (ParamGetBool(m_iObjId, "inv"))
	{
		pFrobMsg.message = "FrobInvEnd";
		pFrobMsg.SrcLoc = kFrobLocInv;
	}
	else
	{
		pFrobMsg.message = "FrobWorldEnd";
		pFrobMsg.SrcLoc = kFrobLocWorld;
	}

	g_pScriptManager->SendMessage(&pFrobMsg, pReply);
	return 0;
}


/*******************
 * TrigFrobber
 */
void cScr_TrigFrobber::InitScriptData()
{
	cScript::InitScriptData();
	if (ParamGetBool(m_iObjId, "inv"))
		RegisterMessageHandler("FrobInvEnd", static_cast<MessageHandlerFunc>(&cScr_TrigFrobber::OnFrob));
	else
		RegisterMessageHandler("FrobWorldEnd", static_cast<MessageHandlerFunc>(&cScr_TrigFrobber::OnFrob));
	return;
}

MSGHANDLER cScr_TrigFrobber::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrMsg msg;
	msg.from = 0; // odd... the Send/PostMessage loses if I use m_iObjId.
	msg.to = m_iObjId;
	msg.time = pMsg->time;
	msg.flags = 8; // And someday I might even learn what this means.

	msg.message = ParamGetString(m_iObjId, "frobmsg", "TurnOn");
	g_pScriptManager->SendMessage(&msg, pReply);

	return 0;
}


/*******************
 * TrapMoveRelative
 */
MSGHANDLER cScr_MoveRelativeTrap::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		FixupPlayerLinks();
	}
	return 1;
}

/*
 * cScrVec.x = Bank;
 * cScrVec.y = Pitch;
 * cScrVec.z = Heading;
 */
struct MoveRelative_Struct
{
	cScrVec	vPosition,
			vFacing,
			vVelocity,
			vRotation;
	bool bRotateAxes;
};

int cScr_MoveRelativeTrap::LinkIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	MoveRelative_Struct* pSDelta = reinterpret_cast<MoveRelative_Struct*>(pData);

	sLink sl;
	pLQ->Link(&sl);
	int iTargetObj = sl.dest;

	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	if (pSDelta->bRotateAxes)
	{
		pObjSrv->Teleport(iTargetObj, pSDelta->vPosition, pSDelta->vFacing, iTargetObj);
	}
	else
	{
		cScrVec vDestPos, vDestFacing;
		pObjSrv->Position(vDestPos,iTargetObj);
		vDestPos += pSDelta->vPosition;
		pObjSrv->Facing(vDestFacing,iTargetObj);
		vDestFacing += pSDelta->vFacing;
		pObjSrv->Teleport(iTargetObj, vDestPos, vDestFacing, 0);
	}

	SInterface<IPropertyManager> pPropMan(g_pScriptManager);
	SInterface<IProperty> pProp;
	pProp = pPropMan->GetPropertyNamed("PhysState");
	if (pProp)
	{
		IGenericProperty* pPhysStateProp;
		sPhysState* pPhysState;
		if (E_NOINTERFACE != pProp->QueryInterface(IID_IPhysStateProperty, reinterpret_cast<void**>(&pPhysStateProp)))
		{
			if (pPhysStateProp->Get(iTargetObj, reinterpret_cast<void**>(&pPhysState)))
			{
				cScrVec v;
				v = pPhysState->vels;
				pPhysState->vels = v + pSDelta->vVelocity;
				v = pPhysState->rot_vels;
				pPhysState->rot_vels = v + pSDelta->vRotation;
				pPhysStateProp->Set(iTargetObj, reinterpret_cast<void*>(pPhysState));
			}
			pPhysStateProp->Release();
		}
	}

	return 1;
}

long cScr_MoveRelativeTrap::OnControlString(const char* pszControl, sScrMsg* pMsg, sMultiParm* pReply)
{
	MoveRelative_Struct vDelta;

	char* pszParams = new char[strlen(pszControl)+1];
	strcpy(pszParams, pszControl);
	char* pszCurrent;
	for (pszCurrent = strtok(pszParams, ";"); pszCurrent; pszCurrent = strtok(NULL, ";"))
	{
		// position/heading or velocity/rotation?
		bool delta;
		if (tolower(*pszCurrent) == 'd')
		{
			delta = TRUE;
			pszCurrent++;
		}
		else
			delta = FALSE;
		// get value
		char* pszValue = strchr(pszCurrent, '=') + 1;
		float value = strtod(pszValue, NULL);
		switch (tolower(*pszCurrent))
		{
		  case 'x':
			if (delta)
				vDelta.vVelocity.x = value;
			else
				vDelta.vPosition.x = value;
			break;
		  case 'y':
			if (delta)
				vDelta.vVelocity.y = value;
			else
				vDelta.vPosition.y = value;
			break;
		  case 'z':
			if (delta)
				vDelta.vVelocity.z = value;
			else
				vDelta.vPosition.z = value;
			break;
		  case 'b':
			if (delta)
				vDelta.vRotation.x = value;
			else
				vDelta.vFacing.x = value;
			break;
		  case 'p':
			if (delta)
				vDelta.vRotation.y = value;
			else
				vDelta.vFacing.y = value;
			break;
		  case 'h':
			if (delta)
				vDelta.vRotation.z = value;
			else
				vDelta.vFacing.z = value;
			break;
		}
	}
	delete[] pszParams;

	vDelta.bRotateAxes = ParamGetBool(m_iObjId, "rotate_axes");
	IterateLinks("ControlDevice", m_iObjId, 0, LinkIterFunc, this, reinterpret_cast<void*>(&vDelta));

	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_MoveRelativeTrap::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	MoveRelative_Struct vDelta;

	vDelta.vPosition.x = ParamGetFloat(m_iObjId, "x");
	vDelta.vPosition.y = ParamGetFloat(m_iObjId, "y");
	vDelta.vPosition.z = ParamGetFloat(m_iObjId, "z");
	vDelta.vVelocity.x = ParamGetFloat(m_iObjId, "dx");
	vDelta.vVelocity.y = ParamGetFloat(m_iObjId, "dy");
	vDelta.vVelocity.z = ParamGetFloat(m_iObjId, "dz");
	vDelta.vFacing.x = ParamGetFloat(m_iObjId, "b");
	vDelta.vFacing.y = ParamGetFloat(m_iObjId, "p");
	vDelta.vFacing.z = ParamGetFloat(m_iObjId, "h");
	vDelta.vRotation.x = ParamGetFloat(m_iObjId, "db");
	vDelta.vRotation.y = ParamGetFloat(m_iObjId, "dp");
	vDelta.vRotation.z = ParamGetFloat(m_iObjId, "dh");
	vDelta.bRotateAxes = ParamGetBool(m_iObjId, "rotate_axes");
	IterateLinks("ControlDevice", m_iObjId, 0, LinkIterFunc, this, reinterpret_cast<void*>(&vDelta));

	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_MoveRelativeTrap::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*******************
 * TrapCampaignQVar
 */
MSGHANDLER cScr_CampaignQVarTrap::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (! static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		char* pszName = GetQVar(NULL, NULL);
		if (pszName)
		{
			SService<IQuestSrv> pQSrv(g_pScriptManager);
			if (pQSrv->Exists(pszName))
			{
				int iCurrent = pQSrv->Get(pszName);
				pQSrv->Delete(pszName);
				pQSrv->Set(pszName, iCurrent, 1);
			}
			delete[] pszName;
		}
	}
	return 1;
}

char* cScr_CampaignQVarTrap::GetQVar(char* op, int* val)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (!pPropSrv->Possessed(m_iObjId,"TrapQVar"))
		return NULL;
	cMultiParm mpQVar;
	pPropSrv->Get(mpQVar, m_iObjId, "TrapQVar", NULL);
	if (mpQVar.type != kMT_String || static_cast<const char*>(mpQVar) == NULL)
		return NULL;
	
	char* pszProp = mpQVar.psz;
	if (op)
		*op = *pszProp;
	if (val)
		*val = strtol(pszProp+1, NULL, 10);
	char* pszName = strchr(pszProp, ':');
	if (!pszName)
		return NULL;
	char* pszDup = new char[strlen(++pszName)+1];
	strcpy(pszDup, pszName);
	return pszDup;
}

long cScr_CampaignQVarTrap::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	char cOperation;
	int iValue = 0;
	int iCurrent = 0;
	char *pszName = GetQVar(&cOperation, &iValue);
	if (!pszName)
		return 1;
	if (!bTurnOn)
		cOperation |= 0x80;		// flag high-bit for inverse operation
	SService<IQuestSrv> pQSrv(g_pScriptManager);
	if (pQSrv->Exists(pszName))
	{
		iCurrent = pQSrv->Get(pszName);
		//pQSrv->Delete(pszName);	// db management done in Sim now
	}
	switch (cOperation)
	{
	  case '=':
		iCurrent = iValue;
		break;
	  case '!':
	  case '|':
		iCurrent |= iValue;
		break;
	  case ('!'+0x80):
	  case ('|'+0x80):
		iCurrent &= ~iValue;
		break;
	  case '+':
	  case ('-'+0x80):
		iCurrent += iValue;
		break;
	  case '-':
	  case ('+'+0x80):
		iCurrent -= iValue;
		break;
	  case '*':
	  case ('/'+0x80):
	  case ('%'+0x80):
		iCurrent *= iValue;
		break;
	  case '/':
	  case ('*'+0x80):
		iCurrent /= iValue;
		break;
	  case '%':
		iCurrent %= iValue;
		break;
	  case '{':
	  case ('}'+0x80):
		iCurrent <<= iValue;
		break;
	  case '}':
	  case ('{'+0x80):
		iCurrent >>= iValue;
		break;
	  case '\"':
		iCurrent = (iCurrent * 10) + (iValue % 10);
		break;
	  case '#':
		iCurrent = ((iCurrent * 10) + (iValue % 10)) % 10000;
		break;
	  case ('\"'+0x80):
	  case ('#'+0x80):
		iCurrent /= 10;
		break;
	  case '?':
	  {
		SService<IDataSrv> pDS(g_pScriptManager);
		iCurrent += pDS->RandInt(0,iValue);
		break;
	  }
	  case ('?'+0x80):
	  {
		SService<IDataSrv> pDS(g_pScriptManager);
		iCurrent -= pDS->RandInt(0,iValue);
		break;
	  }
	  case 'd':
	  {
		SService<IDataSrv> pDS(g_pScriptManager);
		iCurrent += pDS->RandInt(1,iValue);
		break;
	  }
	  case ('d'+0x80):
	  {
		SService<IDataSrv> pDS(g_pScriptManager);
		iCurrent -= pDS->RandInt(1,iValue);
		break;
	  }
	}
	pQSrv->Set(pszName, iCurrent, 1);
	delete[] pszName;
	return 0;
}


/*****************
 * TrapKillEvents
 */
long cScr_KillEvents::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	SInterface<IInputBinder> pInput(g_pScriptManager);
	SService<IDarkInvSrv> pInvSrv(g_pScriptManager);
	if (bTurnOn)
	{
		long lEventMask = 0;
		const char* pszParam = ParamGetString(m_iObjId, "event_mask");
		if (pszParam)
		{
			if (*pszParam == '$')
			{
				++pszParam;
				SService<IQuestSrv> pQS(g_pScriptManager);
				if (pQS->Exists(pszParam))
				{
					lEventMask = pQS->Get(pszParam);
				}
			}
			else
			{
				lEventMask = strtol(pszParam, NULL, 2);
			}
		}
		long lEventFlags = pInput->GetValidEvents();
		pInput->SetValidEvents(lEventFlags & lEventMask);
		pInvSrv->CapabilityControl(kDrkInvCapCycle, kDrkInvControlOff);
		pInvSrv->CapabilityControl(kDrkInvCapInvFrob, kDrkInvControlOff);

		if (m_iTiming == 0)
			m_iTiming = 10000;
	}
	else
	{
		pInput->SetValidEvents(0xFFFFFFFF);
		pInvSrv->CapabilityControl(kDrkInvCapCycle, kDrkInvControlOn);
		pInvSrv->CapabilityControl(kDrkInvCapInvFrob, kDrkInvControlOn);
	}
	return 0;
}


/****************
 * TrapCamAttach
 */
long cScr_CamAttach::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (bTurnOn)
	{
		if (ParamGetBool(m_iObjId, "nolens"))
		{
			SService<IDebugScrSrv> pDebugSrv(g_pScriptManager);
			char sz[32];
			sprintf(sz, "cam_attach %d", m_iObjId);
			pDebugSrv->Command(sz, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null);
		}
		else
		{
			SService<ICameraSrv> pCam(g_pScriptManager);
			if (ParamGetBool(m_iObjId, "static"))
				pCam->StaticAttach(m_iObjId);
			else
				pCam->DynamicAttach(m_iObjId);
		}
	}
	else
	{
		SService<ICameraSrv> pCam(g_pScriptManager);
		if (ParamGetBool(m_iObjId, "forcereturn"))
			pCam->ForceCameraReturn();
		else
			pCam->CameraReturn(m_iObjId);
	}
	return 0;
}


/*******************
 * TrapFreezePlayer
 */
long cScr_FreezePlayer::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (! m_iIsFrozen )
	{
		float fValue = ParamGetFloat(m_iObjId, "freeze_speed", 0.0);
		const char* pszName = ParamGetString(m_iObjId, "freeze_name", "Freeze");
		SService<IDarkInvSrv> pDISrv(g_pScriptManager);
		pDISrv->AddSpeedControl(pszName, fValue, fValue);
		m_iIsFrozen = 1;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_FreezePlayer::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iIsFrozen)
	{
		const char* pszName = ParamGetString(m_iObjId, "freeze_name", "Freeze");
		SService<IDarkInvSrv> pDISrv(g_pScriptManager);
		pDISrv->RemoveSpeedControl(pszName);
		m_iIsFrozen = 0;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_FreezePlayer::OnScale(float fValue, sScrMsg* pMsg, sMultiParm* pReply)
{
	const char* pszName = ParamGetString(m_iObjId, "freeze_name", "Freeze");
	SService<IDarkInvSrv> pDISrv(g_pScriptManager);
	if (m_iIsFrozen)
		pDISrv->RemoveSpeedControl(pszName);
	pDISrv->AddSpeedControl(pszName, fValue, fValue);
	m_iIsFrozen = 1;
	return 0;
}


/*******************
 * TrapThreshold
 */
MSGHANDLER cScr_ThresholdRelay::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_iGlobalCounter.Init(ParamGetInt(m_iObjId, "initial"));
	return 1;
}

long cScr_ThresholdRelay::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	int iThreshold = ParamGetInt(m_iObjId, "threshold");
	if (ParamGetBool(m_iObjId, "global") || pMsg->from == 0)
	{
		if (++m_iGlobalCounter == iThreshold)
		{
			m_iGlobalCounter = 0;
			CDSend("TurnOn", m_iObjId);
			return 0;
		}
	}
	else
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		cMultiParm mpCounter;
		link lCounter;
		pLS->GetOne(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
		if (!lCounter)
		{
			pLS->Create(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
			if (!lCounter) // good candidate for throwing an exception here
				return 1;
			mpCounter = ParamGetString(m_iObjId, "initial");
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
		}
		else
		{
			pLTS->LinkGetData(mpCounter, lCounter, NULL);
		}
		int iCounter = static_cast<int>(mpCounter);
		if (++iCounter == iThreshold)
		{
			CDSend("TurnOn", m_iObjId);
			mpCounter = "0";
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
			return 0;
		}
		char psz[16];
		sprintf(psz, "%d", iCounter);
		mpCounter = psz;
		pLTS->LinkSetData(lCounter, NULL, mpCounter);
	}
	m_iFlags &= ~kTrapFlagOnce;
	return 0;
}

long cScr_ThresholdRelay::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	int iFloor = ParamGetInt(m_iObjId, "floor");
	if (ParamGetBool(m_iObjId, "global") || pMsg->from == 0)
	{
		if (m_iGlobalCounter > iFloor)
			--m_iGlobalCounter;
	}
	else
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		cMultiParm mpCounter;
		link lCounter;
		pLS->GetOne(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
		if (!lCounter)
		{
			pLS->Create(lCounter, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, pMsg->from);
			if (!lCounter) // good candidate for throwing an exception here
				return 1;
			mpCounter = ParamGetString(m_iObjId, "floor");
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
		}
		else
		{
			pLTS->LinkGetData(mpCounter, lCounter, NULL);
		}
		int iCounter = static_cast<int>(mpCounter);
		if (iCounter > iFloor)
		{
			char psz[16];
			sprintf(psz, "%d", iCounter - 1);
			mpCounter = psz;
			pLTS->LinkSetData(lCounter, NULL, mpCounter);
		}
	}
	m_iFlags &= ~kTrapFlagOnce;
	return 0;
}


/*******************
 * TrapRandomTimer
 */
// Relay to all CD links.
MSGHANDLER cScr_RandomTimer::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	const char* pszMsgName = static_cast<const char*>(pTimerMsg->name);
	if (stricmp(pszMsgName, "TurnOn") && 
		stricmp(pszMsgName, "TurnOff"))
		return 1;
	
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 0;

	RelayMessage(pszMsgName);
	return 0;
}

// Delay for a random amount of time.
MSGHANDLER cScr_RandomTimer::OnTurnOnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	if (!pLock->IsLocked(m_iObjId))
	{
		unsigned long lFlags = 0;
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
		{
			cMultiParm mpFlags;
			pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
			lFlags = static_cast<int>(mpFlags);
		}
		const char* pszMsgName = pMsg->message;
		if ((lFlags & kTrapFlagNoOn) && !stricmp(pszMsgName, "TurnOn"))
			return 1;
		if ((lFlags & kTrapFlagNoOff) && !stricmp(pszMsgName, "TurnOff"))
			return 1;
		if (lFlags & kTrapFlagInvert)
		{
			if (!stricmp(pszMsgName, "TurnOn"))
				pszMsgName = "TurnOff";
			else
				pszMsgName = "TurnOn";
		}

		int iMinTime = ParamGetTime(m_iObjId, "min_time", 0);
		int iMaxTime = ParamGetTime(m_iObjId, "max_time", 0);

		unsigned long iInterval;
		if (iMinTime >= iMaxTime)
			iInterval = iMaxTime;
		else
		{
			SService<IDataSrv> pDS(g_pScriptManager);
			iInterval = pDS->RandInt(iMinTime, iMaxTime);
		}
		if (iInterval > 0)
			g_pScriptManager->SetTimedMessage2(m_iObjId, pszMsgName, iInterval, kSTM_OneShot, cMultiParm::Undef);
		else
			RelayMessage(pszMsgName);
		if (pReply)
			MultiParmAssignInt(pReply, 1);
	}
	return 0;
}

void cScr_RandomTimer::RelayMessage(const char* pszMsgName)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
		if (static_cast<int>(mpFlags) & kTrapFlagOnce)
		{
			mpFlags = 1;
			pPropSrv->SetSimple(m_iObjId, "Locked", mpFlags);
		}
	}
	CDSend(pszMsgName, m_iObjId);
}


/***********************
 * TrapRepeatingRelay
 */
MSGHANDLER cScr_RepeatingRelay::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		if (ParamGetBool(m_iObjId, "autostart"))
			OnTurnOn(pMsg, pReply, eTrace);
	}
	return 1;
}

// Send TurnOn to all CD links.
// Then decrement the param and, if zero, stop the timer.
MSGHANDLER cScr_RepeatingRelay::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	const char* pszMsgName = static_cast<const char*>(pTimerMsg->name);
	if (stricmp(pszMsgName, "Repeat"))
		return 1;
	
	CDSend("TurnOn", m_iObjId);

	int iCount = m_iRepeatCount;
	if (iCount > 0)
	{
		if (--iCount == 0)
		{
			if (m_hTimerHandle != NULL)
			{
				g_pScriptManager->KillTimedMessage(m_hTimerHandle);
				m_hTimerHandle = NULL;
			}
		}
		else
			m_iRepeatCount = iCount;
	}
	
	return 0;
}

MSGHANDLER cScr_RepeatingRelay::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_hTimerHandle == NULL)
		return 1;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
		if (!(static_cast<int>(mpFlags) & kTrapFlagNoOff))
		{
			g_pScriptManager->KillTimedMessage(m_hTimerHandle);
			m_hTimerHandle = NULL;
		}
	}
	else
	{
		g_pScriptManager->KillTimedMessage(m_hTimerHandle);
		m_hTimerHandle = NULL;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

// Send the message, then set a timer.
// Params are "repeat" and "interval".
MSGHANDLER cScr_RepeatingRelay::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 1;

	int iCount = ParamGetInt(m_iObjId, "repeat", -1);
	int iInterval = ParamGetInt(m_iObjId, "interval", 0);

	if (iInterval <= 0)
		return 1;

	if (iCount > 0)
	{
		if (m_hTimerHandle == NULL)
		{
			m_iRepeatCount = --iCount;
			CDSend("TurnOn", m_iObjId);
			if (iCount)
			{
				m_hTimerHandle = g_pScriptManager->SetTimedMessage2(m_iObjId, "Repeat", iInterval, kSTM_Periodic, cMultiParm::Undef);
			}
		}
		else // just prolong the existing timer
			m_iRepeatCount += iCount;
	}
	else if (iCount < 0)
	{
		if (m_hTimerHandle == NULL)
		{
			m_iRepeatCount = -1;
			CDSend("TurnOn", m_iObjId);
			m_hTimerHandle = g_pScriptManager->SetTimedMessage2(m_iObjId, "Repeat", iInterval, kSTM_Periodic, cMultiParm::Undef);
		}
	}

	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "TrapFlags", NULL);
		if (static_cast<int>(mpFlags) & kTrapFlagOnce)
		{
			mpFlags = 1;
			pPropSrv->SetSimple(m_iObjId, "Locked", mpFlags);
		}
	}

	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*********************
 * TrapRenderFlash
 */
int cScr_RenderFlash::Relink(ILinkSrv* pLS, ILinkToolsSrv* pLTS, long flavor, int source, int dest)
{
	link l;
	pLS->GetOne(l, flavor, source, 0);
	int ret = 0;
	if (l)
	{
		sLink sl;
		pLTS->LinkGet(l, sl);
		ret = sl.dest;
		pLS->Destroy(l);
	}
	link nl;
	// RenderFlash links don't have any data, so no worries.
	pLS->Create(nl, flavor, source, dest);
	return ret;
}

long cScr_RenderFlash::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (!bTurnOn)
		return 0;
	
	object oPlayer = StrToObject("Player");
	// A bit of trickery here.  The renderflash is linked from the archetype of "Player",
	// not the concrete object. 
	SInterface<ITraitManager> pTraits(g_pScriptManager);
	int iAvatar = pTraits->GetArchetype(oPlayer);
	if (iAvatar != 0 && oPlayer != 0)
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		// The render flash that is linked from "Garrett" will be shown.
		long flavor = pLTS->LinkKindNamed("RenderFlash");
		long l = 0;
		sLink sl;
		l = GetAnyLinkInheritedSrc("RenderFlash", m_iObjId, 0, &sl);
		if (l != 0)
		{
			//pLTS->LinkGet(l, sl);
			int oldflash = Relink(pLS, pLTS, flavor, iAvatar, sl.dest);
			SService<ICameraSrv> pCam(g_pScriptManager);
			SService<ISoundScrSrv> pSound(g_pScriptManager);
			true_bool ret;
			int iSchema = GetAnyLinkDest("SoundDescription", m_iObjId);
			if (iSchema)
				pSound->PlaySchemaAmbient(ret, oPlayer, iSchema, kSoundNetwork0);
			pCam->StaticAttach(oPlayer);
			pCam->CameraReturn(oPlayer);
			Relink(pLS, pLTS, flavor, iAvatar, oldflash);
		}
	}
	return 0;
}


/*******************
 * TrapDisarmPlayer
 */
int cScr_DisarmPlayer::DisarmIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	IObjectSrv* pOS = reinterpret_cast<IObjectSrv*>(pData);
	sLink sl;
	pLQ->Link(&sl);
	int iObjId = static_cast<cScr_DisarmPlayer*>(pScript)->m_iObjId;
	bool bPotions = ParamGetBool(iObjId, "nopotions", 1);
	object oWeapons, oProjectiles, oCrystals, oGrenadz, oPotions;
	pOS->Named(oWeapons, "Weapon");
	pOS->Named(oProjectiles, "Projectile");
	pOS->Named(oCrystals, "Crystal");
	pOS->Named(oGrenadz, "Grenadz");
	pOS->Named(oPotions, "Potions");
	true_bool child_p = FALSE;
	if (oWeapons)
		pOS->InheritsFrom(child_p, sl.dest, oWeapons);
	if (!child_p && oProjectiles)
		pOS->InheritsFrom(child_p, sl.dest, oProjectiles);
	if (!child_p && oCrystals)
		pOS->InheritsFrom(child_p, sl.dest, oCrystals);
	if (!child_p && oGrenadz)
		pOS->InheritsFrom(child_p, sl.dest, oGrenadz);
	if (!child_p && bPotions && oPotions)
		pOS->InheritsFrom(child_p, sl.dest, oPotions);
	if (child_p)
	{
		SService<IContainSrv> pContainSrv(g_pScriptManager);
		pContainSrv->Remove(sl.dest, 0);
		pContainSrv->Add(sl.dest, iObjId, 0, 1);
	}
	return 1;
}

long cScr_DisarmPlayer::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	object oPlayer;
	pObjSrv->Named(oPlayer, "Player");
	if (bTurnOn)
	{
		IterateLinks("Contains", oPlayer, 0, DisarmIterFunc, this, reinterpret_cast<void*>(pObjSrv.get()));
	}
	else
	{
		SService<IContainSrv> pContainSrv(g_pScriptManager);
		pContainSrv->MoveAllContents(m_iObjId, oPlayer, 1);
	}
	return 0;
}

MSGHANDLER cScr_DisarmPlayer::OnContainer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sContainerScrMsg*>(pMsg)->event == kContainAdd)
	{
		SService<IPropertySrv> pPS(g_pScriptManager);
		if (!pPS->Possessed(static_cast<sContainerScrMsg*>(pMsg)->containee, "HasRefs"))
			pPS->Add(static_cast<sContainerScrMsg*>(pMsg)->containee, "HasRefs");
		pPS->SetSimple(static_cast<sContainerScrMsg*>(pMsg)->containee, "HasRefs", 0);
		return 0;
	}
	return 1;
}


/*******************
 * ObjFrobInert
 */
int cScr_BaseFrobInert::MakeUnfrobbableIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	sLink sl;
	pLQ->Link(&sl);
	AddSingleMetaProperty("FrobInert", sl.dest);
	return 1;
}

int cScr_BaseFrobInert::MakeFrobbableIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	sLink sl;
	pLQ->Link(&sl);
	RemoveSingleMetaProperty("FrobInert", sl.dest);
	return 1;
}

void cScr_BaseFrobInert::DoFrobInert(bool bFrobbable)
{
	IterateLinksByData("ScriptParams", m_iObjId, 0, "FrobInert", -1, 
			(bFrobbable) ? MakeFrobbableIterFunc : MakeUnfrobbableIterFunc, this, NULL);
}

bool cScr_BaseFrobInert::IsFrobbable(void)
{
	return m_iDead != 0;
}

// Add or Remove the metaproperty based on the current door state.
MSGHANDLER cScr_BaseFrobInert::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		m_iDead.Init(0);
		DoFrobInert(IsFrobbable());
	}	
	return 1;
}

MSGHANDLER cScr_BaseFrobInert::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_iDead = 1;
	DoFrobInert(true);
	return 0;
}

// Force frobbable.
MSGHANDLER cScr_BaseFrobInert::OnDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	DoFrobInert(true);
	if (m_bSim)
		m_iDead = 1;
	return 0;
}



/*******************
 * DoorFrobInert
 */
bool cScr_DoorFrobInert::IsFrobbable(void)
{
	if (m_iDead != 0)
		return true;
	SService<IDoorSrv> pDoor(g_pScriptManager);
	int iState = pDoor->GetDoorState(m_iObjId);
	return iState != kDoorStateClosed;
}

// Door is dead. We may or may not want to terminally remove the metaproperty.
// Check the "breakable" parameter to see if we should treat the door
// as permanently open.
// If a door is destroyed without getting Slain first, there will likely be an
// ugly black rectangle where the door was.
MSGHANDLER cScr_DoorFrobInert::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (ParamGetBool(m_iObjId, "breakable"))
	{
		m_iDead = 1;
		// StdDoor will open the door for us.
	}
	return 0;
}

// Remove the metaproperty.
MSGHANDLER cScr_DoorFrobInert::OnDoorOpening(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	DoFrobInert(true);
	return 0;
}

// Add the metaproperty (if it doesn't already exist).
MSGHANDLER cScr_DoorFrobInert::OnDoorClose(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!m_iDead)
		DoFrobInert(false);
	return 0;
}


/*******************
 * GrabFrobInert
 */
MSGHANDLER cScr_GrabFrobInert::OnFrobWorldEnd(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_iDead = 1;
	DoFrobInert(true);
	return 0;
}


/*******************
 * TweqFrobInert
 */
bool cScr_TweqFrobInert::IsFrobbable(void)
{
	if (m_iDead != 0)
		return true;
	SService<IPropertySrv> pProps(g_pScriptManager);
	if (!pProps->Possessed(m_iObjId, "StTweqJoints"))
		return true;
	
	cMultiParm mpJoint;
	pProps->Get(mpJoint, m_iObjId, "CfgTweqJoints", "Primary Joint");
	int iJoint = static_cast<int>(mpJoint);
	if (iJoint == 0)
		return true;
	char szAnims[12];
	snprintf(szAnims, 12, "Joint%dAnimS", iJoint);
	pProps->Get(mpJoint, m_iObjId, "StTweqJoints", szAnims);
	return (static_cast<int>(mpJoint) & kTweqFlagReverse) != 0;
}

MSGHANDLER cScr_TweqFrobInert::OnTweqComplete(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sTweqMsg* pTweqMsg = static_cast<sTweqMsg*>(pMsg);
	if (pTweqMsg->Op == kTweqOpHaltTweq)
	{
		DoFrobInert(IsFrobbable());
	}
	return 0;
}


/*******************
 * TrapFrobInert
 */
bool cScr_TrapFrobInert::IsFrobbable(void)
{
	if (m_iDead != 0 || m_iFrobbable != 0)
		return true;
	return false;
}

MSGHANDLER cScr_TrapFrobInert::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	int iFlags = 0;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		iFlags = static_cast<int>(mpTrapFlags);
	}
	if (iFlags & kTrapFlagNoOn)
		return 1;
	
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 0;

	if (iFlags & kTrapFlagInvert)
	{
		DoFrobInert(false);
		m_iFrobbable = 0;
	}
	else
	{
		DoFrobInert(true);
		m_iFrobbable = 1;
	}

	int iTiming = 0;
	if (pPropSrv->Possessed(m_iObjId, "ScriptTiming"))
	{
		cMultiParm mpTiming;
		pPropSrv->Get(mpTiming, m_iObjId, "ScriptTiming", NULL);
		iTiming = static_cast<int>(mpTiming);
	}
	if (iTiming > 0)
	{
		g_pScriptManager->SetTimedMessage2(m_iObjId, (iFlags & kTrapFlagInvert) ? "TurnOn" : "TurnOff",
				iTiming, kSTM_OneShot, 0);
	}

	if (iFlags & kTrapFlagOnce)
	{
		pPropSrv->SetSimple(m_iObjId, "Locked", 1);
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

MSGHANDLER cScr_TrapFrobInert::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	int iFlags = 0;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		iFlags = static_cast<int>(mpTrapFlags);
	}
	if (iFlags & kTrapFlagNoOff)
		return 1;
	
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 0;

	if (iFlags & kTrapFlagInvert)
	{
		DoFrobInert(true);
		m_iFrobbable = 1;
	}
	else
	{
		DoFrobInert(false);
		m_iFrobbable = 0;
	}

	if (iFlags & kTrapFlagOnce)
	{
		pPropSrv->SetSimple(m_iObjId, "Locked", 1);
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

MSGHANDLER cScr_TrapFrobInert::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	if (!stricmp(static_cast<const char*>(pTimerMsg->name), "TurnOn"))
	{
		DoFrobInert(true);
		m_iFrobbable = 1;
		return 0;
	}
	else if (!stricmp(static_cast<const char*>(pTimerMsg->name), "TurnOff"))
	{
		DoFrobInert(false);
		m_iFrobbable = 0;
		return 0;
	}
	return 1;
}


/********************
 * UnlinkOnContained
 */
// Remove links from a specified object.
// Only happens once to save time.
MSGHANDLER cScr_ContainedUnlink::OnContained(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sContainedScrMsg*>(pMsg)->event != kContainAdd && 
		static_cast<sContainedScrMsg*>(pMsg)->event != kContainCombine)
		return 1;
	if (!m_bSim || m_iAlreadyUnlinked != 0)
		return 0;

	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	object oAvatar;
	pObjSrv->Named(oAvatar, "Avatar");
	true_bool bIsAvatar;
	if (*(pObjSrv->InheritsFrom(bIsAvatar, static_cast<sContainedScrMsg*>(pMsg)->container, oAvatar)))
	{
		IterateLinksByData("ScriptParams", m_iObjId, 0, "Unlink", -1, ScriptParamsIterFunc, this, NULL);
		IterateLinksByData("~ScriptParams", m_iObjId, 0, "FrobInert", -1, DestroyLinksIterFunc, this, NULL);
		m_iAlreadyUnlinked = 1;
		return 0;
	}
	return 1;
}

int cScr_ContainedUnlink::ScriptParamsIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	sLink sl;
	pLQ->Link(&sl);
	IterateLinks(NULL, sl.dest, sl.source, DestroyLinksIterFunc, pScript, NULL);
	/* DestroyMany doesn't seem to work this way
	char szFrom[12];
	char szTo[12];
	sprintf(szFrom, "%d", sl.dest);
	sprintf(szTo, "%d", sl.source);
	pLS->DestroyMany(0, szFrom, szTo);
	*/
	pLS->Destroy(pLQ->ID());
	return 1;
}

int cScr_ContainedUnlink::DestroyLinksIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	pLS->Destroy(pLQ->ID());
	return 1;
}


/******************
 * TimedGrenade
 */
MSGHANDLER cScr_TimedGrenade::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_iSafety.Init();
	m_hFuseTimer.Init();

	if (m_hFuseTimer)
	{
		RegisterMessageHandler("BashStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
		RegisterMessageHandler("PokeStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
	}
	return 1;
}

MSGHANDLER cScr_TimedGrenade::OnEndScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	return 1;
}

MSGHANDLER cScr_TimedGrenade::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	const char* pszMsgName = static_cast<const char*>(pTimerMsg->name);

	if (!stricmp(pszMsgName, "StackHack"))
	{
		ParamRemove(m_iObjId, "launch_active");
		return 0;
	}
	if (!stricmp(pszMsgName, "GrenadeActive"))
	{
		GrenadeActive(true);
		return 0;
	}
	if (stricmp(pszMsgName, "TimedGrenade"))
		return 1;

	if (m_iSafety)
	{
		m_iSafety = 0;
		m_hFuseTimer = g_pScriptManager->SetTimedMessage2(m_iObjId, "TimedGrenade", 1600, kSTM_OneShot, cMultiParm::Undef);
	}
	else
	{
		m_hFuseTimer = NULL;
		SService<IDamageSrv> pDS(g_pScriptManager);
		pDS->Slay(m_iObjId, 0);
	}
	return 0;
}

MSGHANDLER cScr_TimedGrenade::OnStimulus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	// Intensity is handled by the receptron link.
	//if (static_cast<sStimMsg*>(pMsg)->intensity < 5.0)
	//	return 1;
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	sLink lSource;
	pLTS->LinkGet(static_cast<sStimMsg*>(pMsg)->source, lSource);
	SService<IDamageSrv> pDS(g_pScriptManager);
	pDS->Slay(m_iObjId, lSource.source);
	return 0;
}

MSGHANDLER cScr_TimedGrenade::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_hFuseTimer)
	{
		g_pScriptManager->KillTimedMessage(m_hFuseTimer);
		m_hFuseTimer = NULL;
	}
	SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
	true_bool bRet;
	pSoundSrv->HaltSchema(bRet, m_iObjId, "", 0);
	
	GrenadeFlash();

	return 0;
}

MSGHANDLER cScr_TimedGrenade::OnFrobInv(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StackCount"))
	{
		cMultiParm mpStack;
		pPropSrv->Get(mpStack, m_iObjId, "StackCount", NULL);
		if (static_cast<int>(mpStack) > 1)
		{
			ParamSetInt(m_iObjId, "launch_active", 1);
			// This is gonna get ugly.
			/* so don't do it
			char* pszParams = GetObjectParams(m_iObjId);
			char* pszNew;
			if (pszParams)
			{
				pszNew = new char[strlen(pszParams) + 17];
				sprintf(pszNew, "%s;%s", "launch_active=1", pszParams);
			}
			else
			{
				pszNew = new char[16];
				strcpy(pszNew, "launch_active=1");
			}
			SetObjectParams(m_iObjId, pszNew);
			delete[] pszNew;
			*/
			g_pScriptManager->SetTimedMessage2(m_iObjId, "StackHack", 1, kSTM_OneShot, cMultiParm::Undef);
		}
		else
		{
			GrenadeArmed();
		}
	}
	else
	{
		GrenadeArmed();
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

MSGHANDLER cScr_TimedGrenade::OnFrobWorld(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	if (m_hFuseTimer)
	{
		g_pScriptManager->KillTimedMessage(m_hFuseTimer);
		m_hFuseTimer = NULL;
		if (m_iSafety)
		{
			m_iSafety = 0;
			GrenadeActive(false);
		}
		else
		{
			SService<IDamageSrv> pDS(g_pScriptManager);
			pDS->Slay(m_iObjId, static_cast<sFrobMsg*>(pMsg)->Frobber);
		}
		return 0;
	}
	// Just in case
	return 1;
}

MSGHANDLER cScr_TimedGrenade::OnContained(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!m_bSim)
		return 0;
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	// Activate the schema after being removed from the inventory.
	if (static_cast<sContainedScrMsg*>(pMsg)->event == kContainRemove && m_hFuseTimer)
	{
		GrenadeActive(true);
		return 0;
	}
	else if (static_cast<sContainedScrMsg*>(pMsg)->event == kContainAdd
		  || static_cast<sContainedScrMsg*>(pMsg)->event == kContainCombine)
	{
		ParamRemove(m_iObjId, "launch_active");
	}
	return 1;
}

MSGHANDLER cScr_TimedGrenade::OnCreate(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	// Dynamically created objects don't get a Sim message. Use the message time instead.
	if (pMsg->time > 0 && ParamGetBool(m_iObjId, "launch_active"))
	{
		GrenadeArmed();
		g_pScriptManager->SetTimedMessage2(m_iObjId, "GrenadeActive", 1, kSTM_OneShot, cMultiParm::Undef);
		ParamRemove(m_iObjId, "launch_active");
		return 0;
	}
	return 1;
}

MSGHANDLER cScr_TimedGrenade::OnSchemaDone(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_hFuseTimer) // We're active, so let's play the active loop.
	{
		SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
		true_bool bRet;
		pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event ActiveLoop", m_iObjId, 0, kEnvSoundOnObj, kSoundNetwork0);
		return 0;
	}
	// If not active... who cares?
	return 1;
}

void cScr_TimedGrenade::GrenadeArmed(void)
{
	if (m_hFuseTimer)
	{
		// How in the hell did you get here? Well, let's clean this up.
		g_pScriptManager->KillTimedMessage(m_hFuseTimer);
		m_hFuseTimer = NULL;
	}
	int iFuseTime = 5000;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "ScriptTiming"))
	{
		cMultiParm mpTiming;
		pPropSrv->Get(mpTiming, m_iObjId, "ScriptTiming", NULL);
		iFuseTime = mpTiming;
	}
	if (iFuseTime > 2000)
	{
		m_iSafety = 1;
		iFuseTime -= 1600;
	}
	else // Uh-oh...
	{
		m_iSafety = 0;
	}
	m_hFuseTimer = g_pScriptManager->SetTimedMessage2(m_iObjId, "TimedGrenade", iFuseTime, kSTM_OneShot, cMultiParm::Undef);
}

void cScr_TimedGrenade::GrenadeFlash(void)
{
	SInterface<ITraitManager> pTraits(g_pScriptManager);
	// TriggerWorldFlash ignores the RenderFlash link on the concrete object
	// ... the bastard
	int iObject = pTraits->GetArchetype(m_iObjId);
	if (GetOneLinkInheritedSrc("RenderFlash", iObject, 0, NULL) != 0)
	{
		SService<IDarkPowerupsSrv> pDPS(g_pScriptManager);
		pDPS->TriggerWorldFlash(m_iObjId);
	}
}

void cScr_TimedGrenade::GrenadeActive(bool bActivate)
{

	SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
	true_bool bRet;
	if (bActivate)
	{
		//pSoundSrv->HaltSchema(bRet, m_iObjId, "", 0);
		pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event Activate", m_iObjId, 0, kEnvSoundOnObj, kSoundNetwork0);
		if (!bRet)
		{
			pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event ActiveLoop", m_iObjId, 0, kEnvSoundOnObj, kSoundNetwork0);
		}
	}
	else
	{
		pSoundSrv->HaltSchema(bRet, m_iObjId, "", 0);
		pSoundSrv->PlayEnvSchema(bRet, m_iObjId, "Event Deactivate", m_iObjId, 0, kEnvSoundOnObj, kSoundNetwork0);
	}

	SService<IActReactSrv> pARSrv(g_pScriptManager);
	cMultiParm mpType = kTweqTypeModels;
	cMultiParm mpAction = (bActivate) ? kTweqDoActivate : kTweqDoHalt;
	pARSrv->React(pARSrv->GetReactionNamed("tweq_control"), 1.0, m_iObjId, 0,
			  mpType, mpAction, cMultiParm::Undef, cMultiParm::Undef, cMultiParm::Undef, cMultiParm::Undef, cMultiParm::Undef, cMultiParm::Undef);
	if (!bActivate)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		if (pPropSrv->Possessed(m_iObjId, "CfgTweqModels"))
		{
			cMultiParm mpModel;
			pPropSrv->Get(mpModel, m_iObjId, "CfgTweqModels", "Model 5");
			if (mpModel.type == kMT_String && mpModel.psz != NULL)
			{
				pPropSrv->SetSimple(m_iObjId, "ModelName", mpModel);
			}
		}
	}

	if (bActivate)
	{
		RegisterMessageHandler("BashStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
		RegisterMessageHandler("PokeStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
	}
	else
	{
		UnregisterMessageHandler("BashStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
		UnregisterMessageHandler("PokeStimStimulus", static_cast<MessageHandlerFunc>(&cScr_TimedGrenade::OnStimulus));
	}
}


/****************
 * ZeroGravRoom
 *
 * This script was reverse-engineered from System Shock 2.
 * (In particular, I never would've known to set the base friction.)
 */
MSGHANDLER cScr_ZeroGravRoom::OnEnter(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sRoomMsg* pRMsg = static_cast<sRoomMsg*>(pMsg);
	SService<IPropertySrv> pPSrv(g_pScriptManager);
	pPSrv->Set(pRMsg->MoveObjId, "PhysAttr", "Gravity %", 0.0);
	pPSrv->Set(pRMsg->MoveObjId, "PhysAttr", "Base Friction", 1.0);

	return 0;
}

MSGHANDLER cScr_ZeroGravRoom::OnExit(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sRoomMsg* pRMsg = static_cast<sRoomMsg*>(pMsg);
	SService<IPropertySrv> pPSrv(g_pScriptManager);
	float fGrav = 100.0;
	if (pPSrv->Possessed(pRMsg->ToObjId, "RoomGrav"))
	{
		cMultiParm mpGrav;
		pPSrv->Get(mpGrav, pRMsg->ToObjId, "RoomGrav", NULL);
		fGrav = static_cast<int>(mpGrav);
	}
	pPSrv->Set(pRMsg->MoveObjId, "PhysAttr", "Gravity %", fGrav);

	return 0;
}


/***************
 * TrapFadeOut
 */
long cScr_FadeOut::OnScale(float fValue, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IDarkGameSrv> pDG(g_pScriptManager);
	pDG->FadeToBlack(fValue);
	return 0;
}

long cScr_FadeOut::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IDarkGameSrv> pDG(g_pScriptManager);
	pDG->FadeToBlack(-1);
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*****************
 * LockHelper
 */
MSGHANDLER cScr_LockHelper::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLockSrv(g_pScriptManager);
	if (pLockSrv->IsLocked(m_iObjId))
		IterateLinksByData("ScriptParams", m_iObjId, 0, "FrobLock", -1, LinkIterFunc, this, reinterpret_cast<void*>(1));
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

MSGHANDLER cScr_LockHelper::OnMessage(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	// ILinkSrv::BroadcastOnAllLinks will do this, but it uses a case-sensitive compare.
	// I don't want to be that strict.
	SService<ILockSrv> pLockSrv(g_pScriptManager);
	IterateLinksByData("ScriptParams", m_iObjId, 0, pMsg->message, -1, LinkIterFunc,
			    this, reinterpret_cast<void*>(pLockSrv->IsLocked(m_iObjId)));
	if (!stricmp(pMsg->message, "PickBegin"))
	{
		IterateLinksByData("ScriptParams", m_iObjId, 0, "Picking", -1, LinkIterFunc,
				    this, reinterpret_cast<void*>(pLockSrv->IsLocked(m_iObjId)));
	}
	else if (!stricmp(pMsg->message, "PickEnd"))
	{
		IterateLinksByData("ScriptParams", m_iObjId, 0, "Picking", -1, LinkIterOffFunc,
				    this, reinterpret_cast<void*>(pLockSrv->IsLocked(m_iObjId)));
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

int cScr_LockHelper::LinkIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	int iLocked = reinterpret_cast<int>(pData);
	sLink sl;
	pLQ->Link(&sl);
	SimplePost(sl.source, sl.dest, "TurnOn", iLocked);
	return 1;
}

int cScr_LockHelper::LinkIterOffFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	int iLocked = reinterpret_cast<int>(pData);
	sLink sl;
	pLQ->Link(&sl);
	SimplePost(sl.source, sl.dest, "TurnOff", iLocked);
	return 1;
}


/*****************
 * KeyHelper
 */
MSGHANDLER cScr_KeyHelper::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SimpleSend(m_iObjId, static_cast<sFrobMsg*>(pMsg)->DstObjId, "UseKey");
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/*****************
 * PickHelper
 */
MSGHANDLER cScr_PickHelper::OnFrobInvBegin(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SimpleSend(m_iObjId, static_cast<sFrobMsg*>(pMsg)->DstObjId, "PickBegin");
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

MSGHANDLER cScr_PickHelper::OnFrobInvEnd(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SimpleSend(m_iObjId, static_cast<sFrobMsg*>(pMsg)->DstObjId, "PickEnd");
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/******************
 * TrapSiblingRelay
 */
MSGHANDLER cScr_SiblingRelay::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	int iInit = 0;
	int iLowestSibling = FindFirstObject(m_iObjId);
	m_iBaseObjectId.Init(0);
	m_iCurrentState.Init(ParamGetBool(m_iObjId, "start_on"));
	m_iSwallowTurnOn.Init(0);
	m_iSwallowTurnOff.Init(0);
	if (m_iBaseObjectId == 0)
	{
		m_iBaseObjectId = iLowestSibling;
	}
	else if (m_iBaseObjectId != iLowestSibling)
	{
		// Joy. The base object ID has changed while we weren't looking.
		if (iLowestSibling == m_iObjId)
		{
			sScrDatumTag oldLock = { m_iBaseObjectId,"cScr_SiblingRelay","m_giSiblingLock"};
			if (g_pScriptManager->IsScriptDataSet(&oldLock))
			{
				sMultiParm param;
				param.type = kMT_Undef;
				g_pScriptManager->ClearScriptData(&oldLock, &param);
				iInit = param.i;
			}
		}
		m_iBaseObjectId = iLowestSibling;
	}
	m_giSiblingLock.SetObj(iLowestSibling);
	if (iLowestSibling == m_iObjId)
		m_giSiblingLock.Init(iInit);

	return 1;
}

MSGHANDLER cScr_SiblingRelay::OnDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_iBaseObjectId == m_iObjId)
	{
		int iNewBase = FindFirstObject(0);
		if (iNewBase != 0)
		{
			m_giSiblingLock.Init();
			int iSiblingLock = m_giSiblingLock;
			m_giSiblingLock.SetObj(iNewBase);
			m_giSiblingLock = iSiblingLock;
			SService<IPropertySrv> pPropSrv(g_pScriptManager);
			if (pPropSrv->Possessed(m_iObjId, "Locked"))
				pPropSrv->CopyFrom(iNewBase, "Locked", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "Locked");
			if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
				pPropSrv->CopyFrom(iNewBase, "TrapFlags", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "TrapFlags");
			if (pPropSrv->Possessed(m_iObjId, "ScriptTiming"))
				pPropSrv->CopyFrom(iNewBase, "ScriptTiming", m_iObjId);
			else
				pPropSrv->Remove(iNewBase, "ScriptTiming");
			PropogateMessage("NewBaseObject", iNewBase);
		}
	}
	// Let's go ahead and re-link outside of sim-time.
	// Makes maintenance of networks in DromEd easier.
	int iLeft = 0, iRight = 0;
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed("ScriptParams");
	ILinkQuery* pLQ = pRel->Query(0, m_iObjId);
	for (; !pLQ->Done(); pLQ->Next())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink sl;
			pLQ->Link(&sl);
			iLeft = sl.source;
			break;
		}
	}
	pLQ->Release();
	pLQ = pRel->Query(m_iObjId, 0);
	for (; !pLQ->Done(); pLQ->Next())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink sl;
			pLQ->Link(&sl);
			iRight = sl.dest;
			break;
		}
	}
	pLQ->Release();
	if (iLeft && iRight)
		pRel->AddFull(iLeft, iRight, "Sibling");

	return 0;
}

MSGHANDLER cScr_SiblingRelay::OnNewBaseObject(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (pMsg->data.type == kMT_Int && pMsg->data.i != 0)
	{
		m_iBaseObjectId = pMsg->data.i;
		m_giSiblingLock.SetObj(pMsg->data.i);
	}
	return 0;
}

MSGHANDLER cScr_SiblingRelay::OnNowLocked(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_iBaseObjectId != m_iObjId)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		pPropSrv->SetSimple(static_cast<int>(m_iBaseObjectId), "Locked", 1);
	}
	return 0;
}

MSGHANDLER cScr_SiblingRelay::OnNowUnlocked(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_iBaseObjectId != m_iObjId)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		pPropSrv->SetSimple(static_cast<int>(m_iBaseObjectId), "Locked", 0);
	}
	return 0;
}

void cScr_SiblingRelay::InitTrapVars()
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	int iObj = m_iBaseObjectId;
	if (pPropSrv->Possessed(iObj, "ScriptTiming"))
	{
		cMultiParm mpTiming;
		pPropSrv->Get(mpTiming, iObj, "ScriptTiming", NULL);
		m_iTiming = mpTiming;
	}
	if (pPropSrv->Possessed(iObj, "TrapFlags"))
	{
		cMultiParm mpTrapFlags;
		pPropSrv->Get(mpTrapFlags, iObj, "TrapFlags", NULL);
		m_iFlags = static_cast<int>(mpTrapFlags);
	}
}

// As well as reading from the base object, we always want to allow propogation msgs.
bool cScr_SiblingRelay::IsLocked(sScrMsg* pMsg)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	m_giSiblingLock.Init();
	return m_giSiblingLock == 0 && pLock->IsLocked(static_cast<int>(m_iBaseObjectId));
}

void cScr_SiblingRelay::SetLock(bool bLock)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	pPropSrv->SetSimple(static_cast<int>(m_iBaseObjectId), "Locked", static_cast<int>(bLock));
}

long cScr_SiblingRelay::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iSwallowTurnOn > 0)
	{
		--m_iSwallowTurnOn;
		DebugPrintf("Sib %d ignoring TurnOn", m_iObjId);
	}
	else
	{
		DebugPrintf("Sib %d receiving TurnOn", m_iObjId);
		DebugPrintf("\tCurrent state is %d", int(m_iCurrentState));
		CDSend("TurnOn", m_iObjId);
		m_giSiblingLock.Init();
		if (m_giSiblingLock == 0)
		{
			m_giSiblingLock = 1;
			const char* pszOn = "TurnOn";
			const char* pszOff = "TurnOff";
			if (m_iFlags & kTrapFlagInvert)
			{
				pszOn = "TurnOff";
				pszOff = "TurnOn";
			}
			const char* pszParam = ParamGetString(m_iObjId, "exclusive", "");
			if (!stricmp(pszParam, "on") || !stricmp(pszParam, "all"))
				PropogateMessage(pszOff, 1);
			else
				PropogateMessage(pszOn, 1);
			m_giSiblingLock = 0;
			
			// Must be from Timer
			if (pMsg->from == 0)
				ToggleLevers(true);
		}
		else // m_giSiblingLock
		{
			if (m_iCurrentState != 1)
				ToggleLevers(true);
			m_iTiming = 0;
			m_iFlags = 0;
		}
		m_iCurrentState = 1;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

long cScr_SiblingRelay::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply)
{
	if (m_iSwallowTurnOff > 0)
	{
		--m_iSwallowTurnOff;
		DebugPrintf("Sib %d ignoring TurnOff", m_iObjId);
	}
	else
	{
		DebugPrintf("Sib %d receiving TurnOff", m_iObjId);
		DebugPrintf("\tCurrent state is %d", int(m_iCurrentState));
		CDSend("TurnOff", m_iObjId);
		m_giSiblingLock.Init();
		if (m_giSiblingLock == 0)
		{
			m_giSiblingLock = 1;
			const char* pszOn = "TurnOn";
			const char* pszOff = "TurnOff";
			if (m_iFlags & kTrapFlagInvert)
			{
				pszOn = "TurnOff";
				pszOff = "TurnOn";
			}
			const char* pszParam = ParamGetString(m_iObjId, "exclusive", "");
			if (!stricmp(pszParam, "off") || !stricmp(pszParam, "all"))
				PropogateMessage(pszOn, 1);
			else
				PropogateMessage(pszOff, 1);
			m_giSiblingLock = 0;
			
			// Must be from Timer
			if (pMsg->from == 0)
				ToggleLevers(false);
		}
		else // m_giSiblingLock
		{
			if (m_iCurrentState != 0)
				ToggleLevers(false);
			m_iTiming = 0;
			m_iFlags = 0;
		}
		m_iCurrentState = 0;
	}
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

struct SiblingRelayToggle_Struct
{
	int iState;
	const char* pszMsg;
};

int cScr_SiblingRelay::ToggleIterFunc(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	cScr_SiblingRelay* pRelay = static_cast<cScr_SiblingRelay*>(pScript);
	SiblingRelayToggle_Struct* pToggleData = reinterpret_cast<SiblingRelayToggle_Struct*>(pData);
	if (pToggleData->iState > 0)
		++pRelay->m_iSwallowTurnOn;
	else if (pToggleData->iState == 0)
		++pRelay->m_iSwallowTurnOff;
	sLink sl;
	pLQ->Link(&sl);
	// Oh hell, let's be greedy.
	SimpleSend(pRelay->m_iObjId, sl.dest, pToggleData->pszMsg);
	return 1;
}

// This pisses me off. How friggen' hard is it to synchronize the state of
// a lever without creating a message loop.
void cScr_SiblingRelay::ToggleLevers(bool bState)
{
	SiblingRelayToggle_Struct sToggleData;
	bool bLeverState = bState ^ ParamGetBool(m_iObjId, "invert_levers");
	bLeverState ^= ((m_iFlags & kTrapFlagInvert) == kTrapFlagInvert);
	sToggleData.pszMsg = (bLeverState) ? "TurnOn" : "TurnOff";
	if (ParamGetBool(m_iObjId, "no_swallow"))
		sToggleData.iState = -1;
	else
		sToggleData.iState = bState;
	DebugPrintf("Sib %d setting levers to %s (%d)", m_iObjId, sToggleData.pszMsg, sToggleData.iState);

	IterateLinks("Owns", m_iObjId, 0, ToggleIterFunc, this, reinterpret_cast<void*>(&sToggleData));
}

// Hell, at least this part worked without incident
void cScr_SiblingRelay::PropogateMessage(const char* pszMsg, int param)
{
	DebugPrintf("Sib %d propogating %s (%d)", m_iObjId, pszMsg, param);
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed("ScriptParams");
	ILinkQuery* pLQ = pRel->Query(m_iObjId, 0);
	while (!pLQ->Done())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink slInfo;
			pLQ->Link(&slInfo);
			if (slInfo.dest == m_iObjId)
				break;
			// I think I'll have the originating object send the message.
			// Rather than going peer-to-peer.
			SimpleSend(m_iObjId, slInfo.dest, pszMsg, param);
			pLQ->Release();
			pLQ = pRel->Query(slInfo.dest, 0);
			continue;
		}
		pLQ->Next();
	}
	pLQ->Release();
}

// Traverse the sibling links looking for the lowest object ID.
int cScr_SiblingRelay::FindFirstObject(int iBase)
{
	SInterface<ILinkManager> pLM(g_pScriptManager);
	SInterface<IRelation> pRel;
	pRel = pLM->GetRelationNamed("ScriptParams");
	ILinkQuery* pLQ = pRel->Query(m_iObjId, 0);
	while (!pLQ->Done())
	{
		if (!stricmp(reinterpret_cast<char*>(pLQ->Data()), "Sibling"))
		{
			sLink slInfo;
			pLQ->Link(&slInfo);
			if (slInfo.dest == m_iObjId)
				break;
			if (iBase > slInfo.dest || iBase == 0)
				iBase = slInfo.dest;
			pLQ->Release();
			pLQ = pRel->Query(slInfo.dest, 0);
			continue;
		}
		pLQ->Next();
	}
	pLQ->Release();
	DebugPrintf("Sib %d thinks base object is %d", m_iObjId, iBase);
	return iBase;
}


/******************
 * TrigContact
 */
MSGHANDLER cScr_ContactTrigger::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPhysSrv> pPhysSrv(g_pScriptManager);
	pPhysSrv->SubscribeMsg(cScript::m_iObjId, kPhysContact);
	return 0;
}

MSGHANDLER cScr_ContactTrigger::OnEndScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPhysSrv> pPhysSrv(g_pScriptManager);
	pPhysSrv->UnsubscribeMsg(cScript::m_iObjId, kPhysContact);
	return 0;
}

struct ContactTrigger_Struct
{
	bool bOn;
	int iObject;
};

MSGHANDLER cScr_ContactTrigger::OnPhysContactCreate(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!IsLocked(pMsg))
	{
		ContactTrigger_Struct sContactInfo;
		if (0 == (sContactInfo.iObject = static_cast<sPhysMsg*>(pMsg)->contactObj))
			return 1;
		sContactInfo.bOn = true;
		InitTrigVars();
		if (0 == IterateLinksByData("ScriptParams", cScript::m_iObjId, sContactInfo.iObject, "Contact", -1, FindContactObject, this, reinterpret_cast<void*>(&sContactInfo)))
		{
			IterateLinksByData("ScriptParams", cScript::m_iObjId, 0, "Contact", -1, FindContactArchetype, this, reinterpret_cast<void*>(&sContactInfo));
		}
		return 0;
	}
	return 1;
}

MSGHANDLER cScr_ContactTrigger::OnPhysContactDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!IsLocked(pMsg))
	{
		ContactTrigger_Struct sContactInfo;
		if (0 == (sContactInfo.iObject = static_cast<sPhysMsg*>(pMsg)->contactObj))
			return 1;
		sContactInfo.bOn = false;
		InitTrigVars();
		if (0 == IterateLinksByData("ScriptParams", cScript::m_iObjId, sContactInfo.iObject, "Contact", -1, FindContactObject, this, reinterpret_cast<void*>(&sContactInfo)))
		{
			IterateLinksByData("ScriptParams", cScript::m_iObjId, 0, "Contact", -1, FindContactArchetype, this, reinterpret_cast<void*>(&sContactInfo));
		}
		return 0;
	}
	return 1;
}

int cScr_ContactTrigger::FindContactObject(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	ContactTrigger_Struct* pContactInfo = reinterpret_cast<ContactTrigger_Struct*>(pData);
	cScr_ContactTrigger* pContactTrigger = static_cast<cScr_ContactTrigger*>(pScript);
	if (pContactInfo->bOn)
		pContactTrigger->DoTurnOn(0);
	else
		pContactTrigger->DoTurnOff(0);
	return 0;
}

int cScr_ContactTrigger::FindContactArchetype(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	ContactTrigger_Struct* pContactInfo = reinterpret_cast<ContactTrigger_Struct*>(pData);
	cScr_ContactTrigger* pContactTrigger = static_cast<cScr_ContactTrigger*>(pScript);
	SService<IObjectSrv> pOS(g_pScriptManager);
	sLink sl;
	pLQ->Link(&sl);
	true_bool child_p;
	if (*(pOS->InheritsFrom(child_p, pContactInfo->iObject, sl.dest)))
	{
		if (pContactInfo->bOn)
			pContactTrigger->DoTurnOn(0);
		else
			pContactTrigger->DoTurnOff(0);
		return 0;
	}
	return 1;
}


/*************
 * TrapText
 */
long cScr_TrapText::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (bTurnOn)
	{
		if (ShowBook(m_iObjId, false))
		{
			char* psz = GetObjectParams(m_iObjId);
			if (psz)
			{
				ShowString(psz, CalcTextTime(psz), 0);
				g_pMalloc->Free(psz);
			}
		}
	}
	else
	{
		ShowString("", 1);
	}
	return 0;
}

long cScr_TrapText::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	return OnSwitch(true, pMsg, pReply);
	/*
	sScrMsg MyMsg;
	MyMsg.message = "TurnOn";
	MyMsg.from = pMsg->from;
	MyMsg.to = pMsg->to;
	MyMsg.time = pMsg->time;
	MyMsg.flags = pMsg->flags;
	return ReceiveMessage(&MyMsg, pReply, kNoAction);
	*/
}


/****************
 * FocusPlaque
 */
MSGHANDLER cScr_FocusPlaque::OnFocus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	ShowBook(m_iObjId, false);
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}

MSGHANDLER cScr_FocusPlaque::OnDeFocus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (pReply)
		MultiParmAssignInt(pReply, 1);
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		if ((static_cast<int>(mpTrapFlags) & kTrapFlagNoOff) == 0)
		{
			ShowString("", 1);
			return 0;
		}
	}
	else
	{
			ShowString("", 1);
			return 0;
	}
	return 1;
}


/********************
 * ContainerPhysics
 */
MSGHANDLER cScr_ContainerPhysics::OnSlain(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SInterface<IContainSys> pContainSys(g_pScriptManager);
	sContainIter* pItem = pContainSys->IterStart(m_iObjId);
	if (pItem)
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		SService<IObjectSrv> pObjSrv(g_pScriptManager);
		SService<IPhysSrv> pPhysSrv(g_pScriptManager);
		SService<IDataSrv> pDataSrv(g_pScriptManager);
		cScrVec vDims;
		if (pPropSrv->Possessed(m_iObjId, "PhysType"))
		{
			cMultiParm mpShape;
			pPropSrv->Get(mpShape, m_iObjId, "PhysType", "Type");
			switch (static_cast<int>(mpShape))
			{
			case kPMT_Sphere:
			case kPMT_SphereHat:
				pPropSrv->Get(mpShape, m_iObjId, "PhysDims", "Radius 1");
				vDims.x = static_cast<float>(mpShape);
				vDims.y = static_cast<float>(mpShape);
				vDims.z = static_cast<float>(mpShape);
				break;
			case kPMT_OBB:
				pPropSrv->Get(mpShape, m_iObjId, "PhysDims", "Size");
				vDims.x = static_cast<const mxs_vector*>(mpShape)->x / 2;
				vDims.y = static_cast<const mxs_vector*>(mpShape)->y / 2;
				vDims.z = static_cast<const mxs_vector*>(mpShape)->z / 2;
				break;
			default:
				break;
			}
		}
		cScrVec vVelocity;
		pPhysSrv->GetVelocity(m_iObjId, vVelocity);
		float fAbsorption = (static_cast<sSlayMsg*>(pMsg)->kind == -900) ? -0.1 : 0.6;
		vVelocity.x *= fAbsorption;
		vVelocity.y *= fAbsorption;
		vVelocity.z *= fAbsorption;
		cScrVec vOffset;
		while (1)
		{
			vOffset.x = vDims.x * pDataSrv->RandFltNeg1to1();
			vOffset.y = vDims.y * pDataSrv->RandFltNeg1to1();
			vOffset.z = vDims.z * pDataSrv->RandFltNeg1to1();
			pObjSrv->Teleport(pItem->Object, vOffset, cScrVec::Zero, m_iObjId);
			pContainSys->Remove(pItem->Object, 0);
			pPropSrv->Remove(pItem->Object, "PhysType");
			pPropSrv->Add(pItem->Object, "PhysType");
			pPhysSrv->SetVelocity(pItem->Object, vVelocity);
			if (!pContainSys->IterNext(pItem))
				break;
		}
		pContainSys->IterEnd(pItem);
	}
	return 0;
}


/*******************
 * RandomPatrolPath
 */
MSGHANDLER cScr_RandomPatrolPath::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);

		std::vector<long> vLinks;
		linkset lsLinks;
		pLS->GetAll(lsLinks, pLTS->LinkKindNamed("AICurrentPatrol"), m_iObjId, 0);
		for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
		{
			vLinks.insert(vLinks.begin(), lsLinks.Link());
		}

		if (!vLinks.empty())
		{
			SService<IDataSrv> pDS(g_pScriptManager);
			int r = pDS->RandInt(0, vLinks.size() - 1);
			long dest = vLinks[r];
			vLinks.erase(vLinks.begin() + r);
			std::vector<long>::const_iterator p,q;
			for (p = vLinks.begin(), q = vLinks.end(); p != q; ++p)
			{
				pLS->Destroy(*p);
			}

			if (ParamGetBool(m_iObjId, "teleport_ai"))
			{
				sLink sl;
				pLTS->LinkGet(dest, sl);
				SService<IObjectSrv> pOS(g_pScriptManager);
				cScrVec z;
				pOS->Teleport(m_iObjId, z, z, sl.dest);
			}
		}
	}
	return 1;
}


/***********************
 * RandomStartingPoint
 */
MSGHANDLER cScr_RandomStartingPoint::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		SService<IObjectSrv> pObjSrv(g_pScriptManager);
		object oAvatar;
		pObjSrv->Named(oAvatar, "Avatar");
		true_bool bIsAvatar;
		if (*(pObjSrv->InheritsFrom(bIsAvatar, m_iObjId, oAvatar)))
			return 1;

		int iDiff = -1;
		{
			SService<IQuestSrv> pQvars(g_pScriptManager);
			if (pQvars->Exists("difficulty"))
			{
				iDiff = pQvars->Get("difficulty");
				if (iDiff > 9) iDiff = 9;
			}
		}

		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);

		std::vector<long> vLinks;
		linkset lsLinks;
		pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, 0);
		for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
		{
			char* pLinkData = reinterpret_cast<char*>(lsLinks.Data());
			if (pLinkData && !strnicmp(pLinkData, "start", 5))
			{
				if (iDiff >= 0)
				{
					if (!strchr(pLinkData, '0' + iDiff))
						continue;
				}
				vLinks.insert(vLinks.begin(), lsLinks.Link());
			}
		}

		if (!vLinks.empty())
		{
			SService<IDataSrv> pDS(g_pScriptManager);
			sLink sl;
			pLTS->LinkGet(vLinks.at(pDS->RandInt(0, vLinks.size() - 1)), sl);
			if (sl.dest && sl.dest != m_iObjId)
			{
				pObjSrv->Teleport(m_iObjId, cScrVec::Zero, cScrVec::Zero, sl.dest);
			}
		}
	}
	return 1;
}


/***********************
 * TrigDoorOpening
 */
MSGHANDLER cScr_TrigDoorOpening::OnDoorOpening(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	const char *szMsg = "GoForward";
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		if (static_cast<int>(mpTrapFlags) & kTrapFlagInvert)
			szMsg = "GoReverse";
	}
	CDSend(szMsg, m_iObjId);
	return 0;
}

MSGHANDLER cScr_TrigDoorOpening::OnDoorClosing(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	const char *szMsg = "GoReverse";
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		if (static_cast<int>(mpTrapFlags) & kTrapFlagInvert)
			szMsg = "GoForward";
	}
	CDSend(szMsg, m_iObjId);
	return 0;
}

MSGHANDLER cScr_TrigDoorOpening::OnDoorHalt(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	CDSend("TurnOff", m_iObjId);
	return 0;
}


/***********************
 * InvCarryOver
 */
MSGHANDLER cScr_InvCarryOver::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (! static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		cAnsiStr sPrefix;
		const char* psz = ParamGetString(m_iObjId, "inv_code");
		if (psz)
		{
			sPrefix = psz;
		}
		sPrefix += "inv_";

		SInterface<ITraitManager> pTraits(g_pScriptManager);
		SService<IObjectSrv> pOS(g_pScriptManager);
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		SService<IQuestSrv> pQSrv(g_pScriptManager);

		object oContainer = m_iObjId;
		if (ParamGetBool(m_iObjId, "inv_player"))
		{
			pOS->Named(oContainer, "Player");
			if (!oContainer)
				return 1;
		}
		else
		{
			object oPlayer;
			pOS->Named(oPlayer, "Player");
			if (oPlayer == m_iObjId)
			{
				// Don't do anything if this is the Player object.
				// Leave it up to the starting-point to do the work.
				return 1;
			}
			// See if this is the starting-point marker
			// and redirect to the Player if so.
			true_bool __p;
			pLS->AnyExist(__p, pLTS->LinkKindNamed("PlayerFactory"), m_iObjId, 0);
			if (__p)
			{
				if (oPlayer != 0)
					oContainer = oPlayer;
			}
		}

		SService<IDebugScrSrv> pDS(g_pScriptManager);
		char num[16];
		{
			sprintf(num, " (%d)", static_cast<int>(oContainer));
			cScrStr name;
			pOS->GetName(name, oContainer);
			pDS->MPrint("Carrying Over contents of ", name, num, cScrStr::Null,
						cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null);
		}
		
		std::vector<int> vArchetypes;
		{
			linkset lsLinks;
			pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, 0);
			for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
			{
				char* pLinkData = reinterpret_cast<char*>(lsLinks.Data());
				if (pLinkData && !stricmp(pLinkData, "CarryOver"))
				{
					sLink sl = lsLinks.Get();
					vArchetypes.insert(vArchetypes.end(), sl.dest);
				}
			}
		}

		if (! vArchetypes.empty())
		{
			linkset lsLinks;
			pLS->GetAll(lsLinks, pLTS->LinkKindNamed("Contains"), oContainer, 0);
			for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
			{
				sLink sl = lsLinks.Get();
				for (std::vector<int>::const_iterator arc = vArchetypes.begin(); arc != vArchetypes.end(); ++arc)
				{
					true_bool _p;
					if (*(pOS->InheritsFrom(_p, sl.dest, *arc)))
					{
						object oArc = pTraits->GetArchetype(sl.dest);
						cScrStr sArcName;
						pOS->GetName(sArcName, oArc);
						cAnsiStr sQVar = sPrefix + static_cast<const char*>(sArcName);
						cMultiParm mpCount = 0;
						if (pPropSrv->Possessed(sl.dest, "StackCount"))
							pPropSrv->Get(mpCount, sl.dest, "StackCount", NULL);
						if (pQSrv->Exists(sQVar))
							pQSrv->Delete(sQVar);
						pQSrv->Set(sQVar, mpCount, 1);

						sprintf(num, "%d", static_cast<int>(mpCount));
						pDS->MPrint("  <- ", static_cast<const char*>(sQVar), " = ", num, 
									cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null);
						break;
					}
				}
			}
		}
	}
	return 1;
}


/***********************
 * InvLoadUp
 */
MSGHANDLER cScr_InvLoadUp::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		SService<IPropertySrv> pPS(g_pScriptManager);
		if (pPS->Possessed(m_iObjId, "DiffScript"))
		{
			SService<IQuestSrv> pQS(g_pScriptManager);
			cMultiParm mpProp;
			pPS->Get(mpProp, m_iObjId, "DiffScript", NULL);
			if (!(static_cast<int>(mpProp) & (1 << pQS->Get("difficulty"))))
			{
				ClearLoadUp();
			}
		}
		else
		{
			DoLoadUp();
		}
	}
	return 1;
}

MSGHANDLER cScr_InvLoadUp::OnDifficulty(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	DoLoadUp();
	return 1;
}

void cScr_InvLoadUp::DoLoadUp(void)
{
	cAnsiStr sPrefix;
	const char* psz = ParamGetString(m_iObjId, "inv_code");
	if (psz)
	{
		sPrefix = psz;
	}
	sPrefix += "inv_";

	SInterface<IQuestData> pQuest(g_pScriptManager);
	SService<IContainSrv> pContains(g_pScriptManager);
	SService<IObjectSrv> pOS(g_pScriptManager);
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	
	object oContainer = m_iObjId;
	if (ParamGetBool(m_iObjId, "inv_player"))
	{
		pOS->Named(oContainer, "Player");
		if (!oContainer)
		{
			SService<ILinkSrv> pLS(g_pScriptManager);
			SService<ILinkToolsSrv> pLTS(g_pScriptManager);
			link pfLink;
			pLS->GetOne(pfLink, pLTS->LinkKindNamed("PlayerFactory"), 0, 0);
			if (pfLink == 0)
				return;
			sLink sl;
			pLTS->LinkGet(pfLink, sl);
			oContainer = sl.source;
		}
	}
	else
	{
		object oPlayer;
		pOS->Named(oPlayer, "Player");
		if (oPlayer == m_iObjId)
		{
			// Don't do anything if this is the Player object.
			// Leave it up to the starting-point to do the work.
			return;
		}
		// See if this is the starting-point marker
		// and redirect to the Player if so.
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		true_bool __p;
		pLS->AnyExist(__p, pLTS->LinkKindNamed("PlayerFactory"), m_iObjId, 0);
		if (__p)
		{
			if (oPlayer != 0)
				oContainer = oPlayer;
		}
	}

	SService<IDebugScrSrv> pDS(g_pScriptManager);
	char num[16];
	{
		sprintf(num, " (%d)", static_cast<int>(oContainer));
		cScrStr name;
		pOS->GetName(name, oContainer);
		pDS->MPrint("Loading Up contents of ", name, num, cScrStr::Null,
					cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null);
	}

	IQuestDataIter* qvIter = pQuest->Iter(1);
	for (; !qvIter->Done(); qvIter->Next())
	{
		const char* qvName = qvIter->Name();
		if (!sPrefix.Compare(sPrefix.GetLength(), qvName))
		{
			const char* szArc = qvName + sPrefix.GetLength();
			object oArc, oInv;
			pOS->Named(oArc, szArc);
			if (oArc != 0)
			{
				sprintf(num, "%d", qvIter->Value());
				pDS->MPrint("  -> ", szArc, " = ", num, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null);
				
				pOS->Create(oInv, oArc);
				if (oInv != 0)
				{
					if (qvIter->Value() != 0)
					{
						if (!pPropSrv->Possessed(oInv, "StackCount"))
							pPropSrv->Add(oInv, "StackCount");
						pPropSrv->SetSimple(oInv, "StackCount", qvIter->Value());
					}
					pContains->Add(oInv, oContainer, 0, 1);
				}
			}
			else
			{
				pDS->MPrint("  unknown archetype: ", szArc, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null, cScrStr::Null);
			}
			pQuest->Delete(qvName);
		}
	}
}

void cScr_InvLoadUp::ClearLoadUp(void)
{
	cAnsiStr sPrefix;
	const char* psz = ParamGetString(m_iObjId, "inv_code");
	if (psz)
	{
		sPrefix = psz;
	}
	sPrefix += "inv_";

	SInterface<IQuestData> pQuest(g_pScriptManager);
	IQuestDataIter* qvIter = pQuest->Iter(1);
	for (; !qvIter->Done(); qvIter->Next())
	{
		const char* qvName = qvIter->Name();
		if (!sPrefix.Compare(sPrefix.GetLength(), qvName))
		{
			pQuest->Delete(qvName);
		}
	}
}


/***********************
 * GenericPotion
 */
int cScr_GenericPotion::GetRecipient(void)
{
	int frobber = GetOneLinkByDataDest("~ScriptParams", m_iObjId, "Potion", -1);
	int dest = ParamGetObject(m_iObjId, "potion_dest", frobber, 0);
	if (dest == 0)
		dest = frobber;
	return dest;
}

MSGHANDLER cScr_GenericPotion::OnStartPotion(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bActive == 0)
	{
		int mp = ParamGetObject(m_iObjId, "potion_effect");
		int dest = GetRecipient();
		if (mp != 0 && dest != 0)
		{
			SService<IObjectSrv> pOS(g_pScriptManager);
			pOS->AddMetaProperty(dest, mp);
		}
		m_bActive = 1;
		return 0;
	}
	return 1;
}

MSGHANDLER cScr_GenericPotion::OnDestroy(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bSim && m_bActive != 0)
	{
		int mp = ParamGetObject(m_iObjId, "potion_effect");
		int dest = GetRecipient();
		if (mp != 0 && dest != 0)
		{
			SService<IObjectSrv> pOS(g_pScriptManager);
			pOS->RemoveMetaProperty(dest, mp);
		}
		m_bActive = 0;
		return 0;
	}
	return 1;
}


/***********************
 * InventoryMetaprop
 */
MSGHANDLER cScr_InvMP::OnContained(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!m_bSim)
		return 0;
	int container = static_cast<sContainedScrMsg*>(pMsg)->container;
	SService<IObjectSrv> pOS(g_pScriptManager);
	object oAvatar;
	pOS->Named(oAvatar, "Avatar");
	true_bool bCond;
	if (*(pOS->InheritsFrom(bCond, container, oAvatar)))
	{
		int mp = ParamGetObject(m_iObjId, "inv_effect");
		if (mp != 0)
		{
			if (static_cast<sContainedScrMsg*>(pMsg)->event == kContainAdd || 
				static_cast<sContainedScrMsg*>(pMsg)->event == kContainCombine)
			{
				if (!*(pOS->HasMetaProperty(bCond, container, mp)))
				{
					pOS->AddMetaProperty(container, mp);
				}
			}
			else if (static_cast<sContainedScrMsg*>(pMsg)->event == kContainRemove)
			{
				if (*(pOS->HasMetaProperty(bCond, container, mp)))
				{
					pOS->RemoveMetaProperty(container, mp);
				}
			}
		}
		return 0;
	}
	return 1;
}


/***********************
 * TrapContainer
 */
MSGHANDLER cScr_ContainerTrap::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		FixupPlayerLinks();
	}
	return 1;
}

struct ContainerTrap_Struct
{
	IContainSrv* pCont;
	int recipient;
	long create_link;
};

int cScr_ContainerTrap::ContainObjects(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	ContainerTrap_Struct *cont_info = reinterpret_cast<ContainerTrap_Struct*>(pData);
	cScr_ContainerTrap *pContainerTrap = static_cast<cScr_ContainerTrap*>(pScript);

	sLink sl;
	pLQ->Link(&sl);
	cont_info->pCont->Remove(sl.dest, 0);
	cont_info->pCont->Add(sl.dest, cont_info->recipient, 0, 0);
	if (cont_info->create_link != 0)
	{
		link l;
		pLS->Create(l, cont_info->create_link, pContainerTrap->m_iObjId, sl.dest);
	}
	return 1;
}

long cScr_ContainerTrap::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	int container = GetOneLinkDest("ControlDevice", m_iObjId);
	if (ParamGetBool(m_iObjId, "track_contains"))
	{
		ContainerTrap_Struct cont_info;
		SService<IContainSrv> pContains(g_pScriptManager);
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		cont_info.pCont = pContains;
		if (bTurnOn)
		{
			cont_info.recipient = container;
			cont_info.create_link = pLTS->LinkKindNamed("Owns");
			IterateLinks("Contains", m_iObjId, 0, ContainObjects, this, reinterpret_cast<void*>(&cont_info));
		}
		else
		{
			cont_info.recipient = m_iObjId;
			cont_info.create_link = 0;
			IterateLinks("Owns", m_iObjId, 0, ContainObjects, this, reinterpret_cast<void*>(&cont_info));
			cMultiParm mpObjId = m_iObjId;
			pLS->DestroyMany(pLTS->LinkKindNamed("Owns"), static_cast<const char*>(mpObjId), "0");
		}
	}
	else
	{
		if (container != 0)
		{
			SService<IContainSrv> pContain(g_pScriptManager);
			if (bTurnOn)
			{
				pContain->MoveAllContents(m_iObjId, container, 1);
			}
			else
			{
				pContain->MoveAllContents(container, m_iObjId, 1);
			}
		}
	}
	return 0;
}


/***********************
 * TrapDispenser
 */
MSGHANDLER cScr_DispenserTrap::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		FixupPlayerLinks();
	}
	return 1;
}

long cScr_DispenserTrap::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (!bTurnOn)
		return 0;

	int container = GetOneLinkDest("ControlDevice", m_iObjId);
	if (container != 0)
	{
		int iArc = ParamGetObject(m_iObjId, "dispense");
		if (iArc == 0)
			return 0;

		SService<IObjectSrv> pOS(g_pScriptManager);
		object oEmit;
		pOS->Create(oEmit, iArc);
		if (oEmit)
		{
			SService<IContainSrv> pContains(g_pScriptManager);
			pContains->Add(oEmit, container, 0, 1);
		}
	}
	return 0;
}


/***********************
 * TrigPurchase
 */
MSGHANDLER cScr_PurchaseTrigger::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		g_pScriptManager->SetTimedMessage2(cScript::m_iObjId, "DidHeBuyIt", 100, kSTM_OneShot, 0);
	}
	return 1;
}

MSGHANDLER cScr_PurchaseTrigger::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!stricmp(static_cast<sScrTimerMsg*>(pMsg)->name, "DidHeBuyIt"))
	{
		InitTrigVars();
		m_iFlags &= (kTrapFlagNoOn | kTrapFlagNoOff | kTrapFlagInvert);
		object oAvatar = StrToObject("Player");
		if (oAvatar == 0)
			return 0;
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		true_bool bCond;
		if (*(pLS->AnyExist(bCond, pLTS->LinkKindNamed("Contains"), oAvatar, cScript::m_iObjId)))
		{
			DoTurnOn(0);
		}
		else
		{
			DoTurnOff(0);
		}
		return 0;
	}
	return 1;
}


/***********************
 * SimpleGun
 */
MSGHANDLER cScr_Gun::OnSelect(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IWeaponSrv> pWeapon(g_pScriptManager);
	pWeapon->Equip(m_iObjId, 0);
	return 0;
}

MSGHANDLER cScr_Gun::OnDeSelect(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IWeaponSrv> pWeapon(g_pScriptManager);
	pWeapon->UnEquip(m_iObjId);
	return 0;
}

MSGHANDLER cScr_Gun::OnFrobBegin(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sLink sl;
	link l = GetAnyLinkInheritedSrc("Projectile", m_iObjId, 0, &sl);
	if (l != 0)
	{
		object oProj = sl.dest;
		SService<IPhysSrv> pPhys(g_pScriptManager);
		cScrVec vecVel;
		object oEmit;
		pPhys->LaunchProjectile(oEmit, m_iObjId, oProj, 1.0, 6, vecVel);

		l = GetAnyLinkInheritedSrc("GunFlash", m_iObjId, 0, &sl);
		if (l != 0)
		{
			SService<ILinkSrv> pLS(g_pScriptManager);
			SService<ILinkToolsSrv> pLTS(g_pScriptManager);
			SService<IObjectSrv> pOS(g_pScriptManager);
			object oFlash = 0;
			pOS->BeginCreate(oFlash, sl.dest);
			if (oFlash)
			{
				object oAttach;
				int vhot;
				cMultiParm mpFlashInfo;
				pLTS->LinkGetData(mpFlashInfo, l, "VHot");
				vhot = static_cast<int>(mpFlashInfo);
				pLTS->LinkGetData(mpFlashInfo, l, "Flags");
				pPhys->SetGravity(oFlash, 0);
				if (static_cast<int>(mpFlashInfo) & 1)
					oAttach = oEmit;
				else
					oAttach = m_iObjId;
				pOS->Teleport(oFlash, cScrVec::Zero, cScrVec::Zero, oAttach);
				l = 0;
				pLS->Create(l, pLTS->LinkKindNamed("ParticleAttachement"), oFlash, oAttach);
				if (l)
				{
					pLTS->LinkSetData(l, "Type", 1);
					pLTS->LinkSetData(l, "vhot #", vhot);
				}
				pOS->SetTransience(oFlash, true);
				pOS->EndCreate(oFlash);
			}
		}
	}
	return 0;
}

MSGHANDLER cScr_Gun::OnFrobEnd(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	return 1;
}


/***********************
 * TurnAtWaypoints
 */
bool cScr_TurnVator::AdjustTweq(const mxs_vector& currvec, const mxs_vector& destvec)
{
	int activate = 0;
	mxs_vector cfg;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	const mxs_vector* rate = m_vTweqRate;

	// X-axis
	if (rate->x > 0.0 && fabs(destvec.x - currvec.x) > 0.005)
	{
		double delta = destvec.x - currvec.x;
		if (delta < 0.0)
			delta += 360.0;
		if (delta > 180.0)
		{
			if (destvec.x > currvec.x)
			{
				if (currvec.x < 0.005)
				{
					cfg.x = rate->x;
					cfg.y = destvec.x;
					cfg.z = 360.0f;
					pPropSrv->Set(m_iObjId, "Position", "Bank", 0xFFFF);
				}
				else
				{
					cfg.x = rate->x;
					cfg.y = 0.0f;
					cfg.z = currvec.x;
				}
			}
			else
			{
				cfg.x = rate->x;
				cfg.y = destvec.x;
				cfg.z = currvec.x;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "x rate-low-high", cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 1AnimS", 3);
		}
		else
		{
			if (destvec.x < currvec.x)
			{
				if (currvec.x > 359.99)
				{
					cfg.x = rate->x;
					cfg.y = 0.0f;
					cfg.z = destvec.x;
					pPropSrv->Set(m_iObjId, "Position", "Bank", 0);
				}
				else
				{
					cfg.x = rate->x;
					cfg.y = currvec.x;
					cfg.z = 0.0f;
				}
			}
			else
			{
				cfg.x = rate->x;
				cfg.y = currvec.x;
				cfg.z = destvec.x;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "x rate-low-high", cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 1AnimS", 1);
		}
		activate = 1;
	}
	else
	{
		cfg.x = 0.0f; cfg.y = 0.0f; cfg.z = 0.0f;
		pPropSrv->Set(m_iObjId, "CfgTweqRotate", "x rate-low-high", cfg);
		pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 1AnimS", 0);
	}

	// Y-axis
	if (rate->y > 0.0 && fabs(destvec.y - currvec.y) > 0.005)
	{
		double delta = destvec.y - currvec.y;
		if (delta < 0.0)
			delta += 360.0;
		if (delta > 180.0)
		{
			if (destvec.y > currvec.y)
			{
				if (currvec.y < 0.005)
				{
					cfg.x = rate->y;
					cfg.y = destvec.y;
					cfg.z = 360.0f;
					pPropSrv->Set(m_iObjId, "Position", "Pitch", 0xFFFF);
				}
				else
				{
					cfg.x = rate->y;
					cfg.y = 0.0f;
					cfg.z = currvec.y;
				}
			}
			else
			{
				cfg.x = rate->y;
				cfg.y = destvec.y;
				cfg.z = currvec.y;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "y rate-low-high", cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 2AnimS", 3);
		}
		else
		{
			if (destvec.y < currvec.y)
			{
				if (currvec.y > 359.99)
				{
					cfg.x = rate->y;
					cfg.y = 0.0f;
					cfg.z = destvec.y;
					pPropSrv->Set(m_iObjId, "Position", "Pitch", 0);
				}
				else
				{
					cfg.x = rate->y;
					cfg.y = currvec.y;
					cfg.z = 0.0f;
				}
			}
			else
			{
				cfg.x = rate->y;
				cfg.y = currvec.y;
				cfg.z = destvec.y;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "y rate-low-high", cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 2AnimS", 1);
		}
		activate = 1;
	}
	else
	{
		cfg.x = 0.0f; cfg.y = 0.0f; cfg.z = 0.0f;
		pPropSrv->Set(m_iObjId, "CfgTweqRotate", "y rate-low-high", cfg);
		pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 2AnimS", 0);
	}

	// Z-axis
	if (rate->z > 0.0 && fabs(destvec.z - currvec.z) > 0.005)
	{
		double delta = destvec.z - currvec.z;
		if (delta < 0.0)
			delta += 360.0;
		if (delta > 180.0)
		{
			if (destvec.z > currvec.z)
			{
				if (currvec.z < 0.005)
				{
					cfg.x = rate->z;
					cfg.y = destvec.z;
					cfg.z = 360.0f;
					pPropSrv->Set(m_iObjId, "Position", "Heading", 0xFFFF);
				}
				else
				{
					cfg.x = rate->z;
					cfg.y = 0.0f;
					cfg.z = currvec.z;
				}
			}
			else
			{
				cfg.x = rate->z;
				cfg.y = destvec.z;
				cfg.z = currvec.z;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "z rate-low-high", cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 3AnimS", 3);
		}
		else
		{
			if (destvec.z < currvec.z)
			{
				if (currvec.z > 359.99)
				{
					cfg.x = rate->z;
					cfg.y = 0.0f;
					cfg.z = destvec.z;
					pPropSrv->Set(m_iObjId, "Position", "Heading", 0);
				}
				else
				{
					cfg.x = rate->z;
					cfg.y = currvec.z;
					cfg.z = 0.0f;
				}
			}
			else
			{
				cfg.x = rate->z;
				cfg.y = currvec.z;
				cfg.z = destvec.z;
			}
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "z rate-low-high", cfg);
			pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 3AnimS", 1);
		}
		activate = 1;
	}
	else
	{
		cfg.x = 0.0f; cfg.y = 0.0f; cfg.z = 0.0f;
		pPropSrv->Set(m_iObjId, "CfgTweqRotate", "z rate-low-high", cfg);
		pPropSrv->Set(m_iObjId, "StTweqRotate", "Axis 3AnimS", 0);
	}

	pPropSrv->Set(m_iObjId, "StTweqRotate", "AnimS", activate);
	return activate;
}

MSGHANDLER cScr_TurnVator::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	mxs_vector vRate = {0.0, 0.0, 0.0};
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StTweqRotate"))
	{
		cMultiParm mpCfg;
		pPropSrv->Get(mpCfg, m_iObjId, "CfgTweqRotate", "x rate-low-high");
		vRate.x = static_cast<const mxs_vector*>(mpCfg)->x;
		pPropSrv->Get(mpCfg, m_iObjId, "CfgTweqRotate", "y rate-low-high");
		vRate.y = static_cast<const mxs_vector*>(mpCfg)->x;
		pPropSrv->Get(mpCfg, m_iObjId, "CfgTweqRotate", "z rate-low-high");
		vRate.z = static_cast<const mxs_vector*>(mpCfg)->x;
		
		pPropSrv->Get(mpCfg, m_iObjId, "CfgTweqRotate", "Primary Axis");
		if (static_cast<int>(mpCfg) != 0)
		{
			mpCfg = 0;
			pPropSrv->Set(m_iObjId, "CfgTweqRotate", "Primary Axis", mpCfg);
		}
	}
	m_vTweqRate.Init(&vRate);
	return 1;
}

MSGHANDLER cScr_TurnVator::OnStart(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StTweqRotate"))
	{
		int x,y,z;
		cMultiParm mpFlags;
		pPropSrv->Get(mpFlags, m_iObjId, "StTweqRotate", "Axis 1AnimS");
		x = static_cast<int>(mpFlags);
		pPropSrv->Get(mpFlags, m_iObjId, "StTweqRotate", "Axis 2AnimS");
		y = static_cast<int>(mpFlags);
		pPropSrv->Get(mpFlags, m_iObjId, "StTweqRotate", "Axis 3AnimS");
		z = static_cast<int>(mpFlags);
		if (x != 0 || y != 0 || z != 0)
		{
			mpFlags = 1;
			pPropSrv->Set(m_iObjId, "StTweqRotate", "AnimS", mpFlags);
		}
	}
	return 0;
}

MSGHANDLER cScr_TurnVator::OnStop(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StTweqRotate"))
	{
		pPropSrv->Set(m_iObjId, "StTweqRotate", "AnimS", 0);
	}
	return 0;
}

MSGHANDLER cScr_TurnVator::OnWaypoint(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "StTweqRotate"))
	{
		pPropSrv->Set(m_iObjId, "StTweqRotate", "AnimS", 0);

		SService<IObjectSrv> pOS(g_pScriptManager);
		cScrVec currvec, destvec;
		pOS->Facing(currvec, m_iObjId);
		pOS->Facing(destvec, static_cast<sMovingTerrainMsg*>(pMsg)->waypoint);
		if (AdjustTweq(currvec, destvec))
		{
			sMultiParm mpDestVec;
			mpDestVec.type = kMT_Vector;
			mpDestVec.pVector = &destvec;
			sScrDatumTag tagDestVec;
			tagDestVec.objId = m_iObjId;
			tagDestVec.pszClass = "cScr_TurnVator";
			tagDestVec.pszName = "m_vDestVec";
			g_pScriptManager->SetScriptData(&tagDestVec, &mpDestVec);
		}
	}
	return 0;
}

MSGHANDLER cScr_TurnVator::OnTweqComplete(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sTweqMsg* pTweqMsg = static_cast<sTweqMsg*>(pMsg);
	if (pTweqMsg->Type == kTweqTypeRotate && pTweqMsg->Op == kTweqOpHaltTweq)
	{
		sScrDatumTag tagDestVec;
		tagDestVec.objId = m_iObjId;
		tagDestVec.pszClass = "cScr_TurnVator";
		tagDestVec.pszName = "m_vDestVec";
		if (g_pScriptManager->IsScriptDataSet(&tagDestVec))
		{
			sMultiParm mpDestVec;
			mpDestVec.type = kMT_Undef;
			g_pScriptManager->GetScriptData(&tagDestVec, &mpDestVec);

			SService<IObjectSrv> pOS(g_pScriptManager);
			cScrVec currvec;
			pOS->Facing(currvec, m_iObjId);
			if (! AdjustTweq(currvec, *mpDestVec.pVector))
			{
				sMultiParm mpRet;
				mpRet.type = kMT_Undef;
				g_pScriptManager->ClearScriptData(&tagDestVec, &mpRet);
				g_pMalloc->Free(mpRet.pVector);
			}
			g_pMalloc->Free(mpDestVec.pVector);
		}
	}
	return 0;
}


/***********************
 * VisibleRelay
 */
bool cScr_VisibleTrap::TestVisible(bool bVisible)
{
	int iObjId = m_iObjId;
	sLink slVisLink;
	if (GetOneLinkByData("ScriptParams", m_iObjId, 0, &slVisLink, "Visible", -1))
	{
		iObjId = slVisLink.dest;
	}
	true_bool bIsVisible;
	SService<IObjectSrv> pObjSrv(g_pScriptManager);
	pObjSrv->RenderedThisFrame(bIsVisible, iObjId);
	DebugPrintf("Object %d is %svisible", iObjId, (bIsVisible)?"":"not ");
	return bVisible == bIsVisible;
}

MSGHANDLER cScr_VisibleTrap::OnQuery(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!pReply)
		return 1;

	const char* pszParm = static_cast<const char*>(pMsg->data);
	if (!pszParm || !*pszParm)
		return 1;

	bool bVisible = true;
	char* pszEnd;
	long iConv = strtol(pszParm, &pszEnd, 10);
	if (pszEnd > pMsg->data.psz)
	{
		switch (iConv)
		{
		case 0:
			bVisible = false; break;
		case 1:
			bVisible = true; break;
		default:
			return 1;
		}
	}
	else
	{
		if (!stricmp(pMsg->data.psz, "visible") || 
			!stricmp(pMsg->data.psz, "on") ||
			!stricmp(pMsg->data.psz, "true") ||
			!stricmp(pMsg->data.psz, "yes"))
		{
			bVisible = true;
		}
		else if (!stricmp(pMsg->data.psz, "invisible") || 
			!stricmp(pMsg->data.psz, "off") ||
			!stricmp(pMsg->data.psz, "false") ||
			!stricmp(pMsg->data.psz, "no"))
		{
			bVisible = false;
		}
		else
			return 1;
	}

	MultiParmAssignInt(pReply, (TestVisible(bVisible)) ? 1 : 0);
	return 0;
}

MSGHANDLER cScr_VisibleTrap::OnTrigger(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILockSrv> pLock(g_pScriptManager);
	if (pLock->IsLocked(m_iObjId))
		return 1;

	int iFlags = 0;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "TrapFlags"))
	{
		cMultiParm mpTrapFlags;
		pPropSrv->Get(mpTrapFlags, m_iObjId, "TrapFlags", NULL);
		iFlags = static_cast<int>(mpTrapFlags);
	}

	if (TestVisible((iFlags & kTrapFlagInvert) == 0))
	{
		CDSend(pMsg->message, m_iObjId, pMsg->data);
	}
	
	if (iFlags & kTrapFlagOnce)
	{
		pPropSrv->SetSimple(m_iObjId, "Locked", 1);
	}

	if (pReply)
		MultiParmAssignInt(pReply, 1);
	return 0;
}


/***********************
 * ResetTweqState
 */
void cScr_ResetTweqs::ResetTweq(const char* pszName, int iObj, int iArc, IPropertySrv* pPS)
{
	if (pPS->Possessed(iObj, pszName) && pPS->Possessed(iArc, pszName))
	{
		cMultiParm mpTime, mpFrame;
		pPS->Get(mpTime, iArc, pszName, "Cur Time");
		pPS->Get(mpFrame, iArc, pszName, "Frame #");
		pPS->Set(iObj, pszName, "Cur Time", mpTime);
		pPS->Set(iObj, pszName, "Frame #", mpFrame);
	}
}

void cScr_ResetTweqs::ResetProp(const char* pszName, int iObj, int iArc, IPropertySrv* pPS)
{
	if (pPS->Possessed(iObj, pszName) && pPS->Possessed(iArc, pszName))
	{
		pPS->Remove(iObj, pszName);
		pPS->Add(iObj, pszName);
	}
}

void cScr_ResetTweqs::ResetAll(int iObj, int iArc, IPropertySrv* pPS)
{
	if (pPS->Possessed(iObj, "StTweqJoints"))
		ResetProp("JointPos", iObj, iArc, pPS);
	ResetTweq("StTweqBlink", iObj, iArc, pPS);
	ResetTweq("StTweqEmit", iObj, iArc, pPS);
	ResetTweq("StTweq2Emit", iObj, iArc, pPS);
	ResetTweq("StTweq3Emit", iObj, iArc, pPS);
	ResetTweq("StTweq4Emit", iObj, iArc, pPS);
	ResetTweq("StTweq5Emit", iObj, iArc, pPS);
	if (pPS->Possessed(iObj, "StTweqModels"))
		ResetProp("ModelName", iObj, iArc, pPS);
	if (pPS->Possessed(iObj, "StTweqScale"))
		ResetProp("Scale", iObj, iArc, pPS);
}

MSGHANDLER cScr_ResetTweqs::OnTweqComplete(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sTweqMsg* pTweqMsg = static_cast<sTweqMsg*>(pMsg);
	SService<IPropertySrv> pPS(g_pScriptManager);
	SInterface<ITraitManager> pTM(g_pScriptManager);
	int iArc = pTM->GetArchetype(m_iObjId);

	if (pTweqMsg->Op == kTweqOpHaltTweq)
	{
		switch (pTweqMsg->Type)
		{
		case kTweqTypeJoints:
			ResetProp("JointPos", m_iObjId, iArc, pPS);
			break;
		case kTweqTypeEmitter:
			ResetTweq("StTweqEmit", m_iObjId, iArc, pPS);
			ResetTweq("StTweq2Emit", m_iObjId, iArc, pPS);
			ResetTweq("StTweq3Emit", m_iObjId, iArc, pPS);
			ResetTweq("StTweq4Emit", m_iObjId, iArc, pPS);
			ResetTweq("StTweq5Emit", m_iObjId, iArc, pPS);
			break;
		case kTweqTypeFlicker:
			ResetTweq("StTweqBlink", m_iObjId, iArc, pPS);
			break;
		case kTweqTypeModels:
			ResetProp("ModelName", m_iObjId, iArc, pPS);
			break;
		case kTweqTypeScale:
			ResetProp("Scale", m_iObjId, iArc, pPS);
			break;
		case kTweqTypeAll:
			ResetAll(m_iObjId, iArc, pPS);
		case kTweqTypeRotate:
			if (pPS->Possessed(m_iObjId, "StTweqRotate"))
			{
				SService<IObjectSrv> pOS(g_pScriptManager);
				cScrVec vFacing = *m_vRotInit;
				cScrVec vPos;
				pOS->Position(vPos, m_iObjId);
				pOS->Teleport(m_iObjId, vPos, vFacing, 0);
			}
			break;
		case kTweqTypeDelete:
		case kTweqTypeLock:
		default:
			break;
		}
	}

	return 0;
}

MSGHANDLER cScr_ResetTweqs::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	SInterface<ITraitManager> pTM(g_pScriptManager);
	int iArc = pTM->GetArchetype(m_iObjId);
	ResetAll(m_iObjId, iArc, pPS);
	if (pPS->Possessed(m_iObjId, "StTweqRotate"))
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		cScrVec vFacing = *m_vRotInit;
		cScrVec vPos;
		pOS->Position(vPos, m_iObjId);
		pOS->Teleport(m_iObjId, vPos, vFacing, 0);
	}
	return 0;
}

MSGHANDLER cScr_ResetTweqs::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		cScrVec vFacing;
		pOS->Facing(vFacing, m_iObjId);
		m_vRotInit.Init(&vFacing);
	}
	return 1;
}


/***********************
 * RecallDevice
 */
int cScr_RecallDevice::RecallIter(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	int iAvatar = reinterpret_cast<int>(pData);
	SService<IObjectSrv> pOS(g_pScriptManager);
	sLink sl;
	pLQ->Link(&sl);
	pOS->Teleport(sl.dest, cScrVec::Zero, cScrVec::Zero, iAvatar);
	return 1;
}

long cScr_RecallDevice::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (bTurnOn)
	{
		int iAvatar = StrToObject("Player");
		IterateLinks("ControlDevice", m_iObjId, 0, RecallIter, this, reinterpret_cast<void*>(iAvatar));
	}
	return 0;
}


/***********************
 * TrigProxFrob
 */
MSGHANDLER cScr_ProxFrobTrigger::OnFrob(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (cScrTrigger::IsLocked(pMsg))
		return 1;
	int iFrobber = static_cast<sFrobMsg*>(pMsg)->Frobber;
	if (!iFrobber)
		return 1;
	int iProx = GetOneLinkDest("Route", cScript::m_iObjId);
	if (!iProx)
		return 1;
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	true_bool bNearby;
	pLS->AnyExist(bNearby, pLTS->LinkKindNamed("Population"), iProx, iFrobber);
	if (bNearby)
	{
		InitTrigVars();
		DoTurnOn(0);
		SService<ISoundScrSrv> pSoundSrv(g_pScriptManager);
		eEnvSoundLoc opt = (iFrobber == StrToObject("Player")) ? kEnvSoundAmbient : kEnvSoundOnObj;
		true_bool _p;
		pSoundSrv->PlayEnvSchema(_p, cScript::m_iObjId, "Event Activate", cScript::m_iObjId, 0, opt, kSoundNetwork0);
	}

	return 0;
}


/***********************
 * PoliteVOTrap
 */
long cScr_VOPolite::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (bTurnOn)
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);

		true_bool bInvested;
		pLS->AnyExist(bInvested, pLTS->LinkKindNamed("~AIInvest"), StrToObject("Player"), 0);
		if (bInvested)
			return 1;

		linkset lsSounds;
		pLS->GetAll(lsSounds, pLTS->LinkKindNamed("SoundDescription"), m_iObjId, 0);
		for (; lsSounds.AnyLinksLeft(); lsSounds.NextLink())
		{
			SService<ISoundScrSrv> pSounds(g_pScriptManager);
			sLink sl = lsSounds.Get();
			true_bool ret;
			pSounds->PlayVoiceOver(ret, m_iObjId, sl.dest);
		}
	}
	return 0;
}


/***********************
 * TrigSchemaDoneSpec
 */
MSGHANDLER cScr_SpecSchemaTrigger::OnSchemaDone(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (cScrTrigger::IsLocked(pMsg))
		return 1;

	const char* pszSchema = ParamGetString(cScript::m_iObjId, "schema");
	if (pszSchema)
	{
		if (!stricmp(pszSchema, static_cast<sSchemaDoneMsg*>(pMsg)->name))
		{
			InitTrigVars();
			DoTurnOn(0);
		}
	}

	return 0;
}


/***********************
 * AIGoMissing
 */
int cScr_MissingPerson::CreateMissingMarker(void)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	object oMark, oArc;
	pOS->Named(oArc, "MissingPerson");
	if (!oArc)
		return 0;
	cScrVec vecOrigin;
	SService<IPropertySrv> pPS(g_pScriptManager);
	if (pPS->Possessed(m_iObjId, "AI_IdleOrgn"))
	{
		cMultiParm mpProp;
		pPS->Get(mpProp, m_iObjId, "AI_IdleOrgn", "Original Location");
		vecOrigin = *static_cast<const mxs_vector*>(mpProp);
	}
	else
	{
		pOS->Position(vecOrigin, m_iObjId);
	}
	
	pOS->Create(oMark, oArc);
	if (oMark)
	{
		pOS->Teleport(oMark, vecOrigin, cScrVec::Zero, 0);
	/*
		pPS->Add(oMark, "SuspObj");
		pPS->Set(oMark, "SuspObj", "Is Suspicious", 1);
	*/
	}
	return oMark;
}

#define DISTANCE2TIME	5000.0
int cScr_MissingPerson::DistanceToTime(int iFrom, int iTo)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	cScrVec vecFrom, vecTo;
	pOS->Position(vecFrom, iFrom);
	pOS->Position(vecTo, iTo);
	vecTo -= vecFrom;
	return int(DISTANCE2TIME * vecTo.Magnitude());
}

void cScr_MissingPerson::PatrolMissing(int iTrolPt)
{
	SimpleSend(m_iObjId, iTrolPt, "GoMissing");

	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	link lPatrol;
	sLink slPatrol;
	pLS->GetOne(lPatrol, pLTS->LinkKindNamed("AIPatrol"), iTrolPt, 0);
	if (!lPatrol)
		return;
	pLTS->LinkGet(lPatrol, slPatrol);
	if (slPatrol.dest != m_iMarkerObj)
	{
		g_pScriptManager->SetTimedMessage2(m_iObjId, "PatrolMissing", DistanceToTime(iTrolPt,slPatrol.dest), kSTM_OneShot, slPatrol.dest);
	}
}

int cScr_MissingPerson::GoMissing(void)
{
	int iRet = 0;
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	true_bool bPatrol;
	pLS->AnyExist(bPatrol, pLTS->LinkKindNamed("AICurrentPatrol"), m_iObjId, 0);
	if (bPatrol)
	{
		link lPatrol;
		sLink slPatrol;
		pLS->GetOne(lPatrol, pLTS->LinkKindNamed("AICurrentPatrol"), m_iObjId, 0);
		pLTS->LinkGet(lPatrol, slPatrol);
		m_iMarkerObj = slPatrol.dest;
		iRet = slPatrol.dest;
		if (ParamGetBool(m_iObjId, "missingpatrol"))
			PatrolMissing(slPatrol.dest);
		else
			SimpleSend(m_iObjId, slPatrol.dest, "GoMissing");
	}
	else
	{
		iRet = CreateMissingMarker();
		if (iRet)
			SimpleSend(m_iObjId, iRet, "GoMissing");
		m_iMarkerObj = iRet;
	}
	return iRet;
}

void cScr_MissingPerson::NotMissing(int iMarker)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	true_bool bPatrol;
	pLS->AnyExist(bPatrol, pLTS->LinkKindNamed("AICurrentPatrol"), m_iObjId, 0);
	if (bPatrol)
	{
		SimpleSend(m_iObjId, iMarker, "NotMissing");

		if (ParamGetBool(m_iObjId, "missingpatrol"))
		{
			linkkind aipatrol = pLTS->LinkKindNamed("AIPatrol");
			link lPatrol;
			sLink slPatrol;
			slPatrol.dest = iMarker;
			do
			{
				pLS->GetOne(lPatrol, aipatrol, slPatrol.dest, 0);
				if (!lPatrol)
					break;
				pLTS->LinkGet(lPatrol, slPatrol);
				SimpleSend(m_iObjId, slPatrol.dest, "NotMissing");
			}
			while (slPatrol.dest != iMarker);
		}

		// Use the link-kill option of AIWatchObj
		//link lMark;
		//pLS->GetOne(lMark, pLTS->LinkKindNamed("AIWatchObj"), m_iObjId, iMarker);
		//if (lMark)
		//	pLS->Destroy(lMark);
	}
	else
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		pOS->Destroy(iMarker);
	}
}

MSGHANDLER cScr_MissingPerson::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!stricmp(static_cast<sScrTimerMsg*>(pMsg)->name, "PatrolMissing"))
	{
		PatrolMissing(static_cast<int>(pMsg->data));
		return 0;
	}
	return 1;
}

MSGHANDLER cScr_MissingPerson::OnModeChange(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sAIModeChangeMsg*>(pMsg)->mode == kAIM_Dead)
	{
		if (m_iMarkerObj == 0)
		{
			GoMissing();
		}
		else
		{
			/* Why?
			SService<ILinkSrv> pLS(g_pScriptManager);
			SService<ILinkToolsSrv> pLTS(g_pScriptManager);
			link lMark;
			pLS->GetOne(lMark, pLTS->LinkKindNamed("AIWatchObj"), m_iObjId, static_cast<int>(m_iMarkerObj));
			if (lMark)
				pLS->Destroy(lMark);
			*/
		}
	}
	return 0;
}

const sAIWatchPoint cScr_MissingPerson::sm_AIWatchLinkDefault = { "",
	kAIWatchSelfEntry, kNoAwareness, kNoVisibility, 0,
	1, 0, 1, 5000, 5000, kNoAlert, kModerateAlert, 
	kAbsolutePriorityResponse, 6, 8,
	{ { kPS_ScrMsg, "AIRelax", "", "", "" }, { kPS_Nothing, "", "", "", "" }, 
	  { kPS_Nothing, "", "", "", "" }, { kPS_Nothing, "", "", "", "" }, 
	  { kPS_Nothing, "", "", "", "" }, { kPS_Nothing, "", "", "", "" }, 
	  { kPS_Nothing, "", "", "", "" }, { kPS_Nothing, "", "", "", "" } }, 
	{ { kPS_Nothing, "", "", "", "" }, { kPS_Nothing, "", "", "", "" }, 
	  { kPS_Nothing, "", "", "", "" }, { kPS_Nothing, "", "", "", "" }, 
	  { kPS_Nothing, "", "", "", "" }, { kPS_Nothing, "", "", "", "" }, 
	  { kPS_Nothing, "", "", "", "" }, { kPS_Nothing, "", "", "", "" } }
};


MSGHANDLER cScr_MissingPerson::OnAlertness(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sAIAlertnessMsg*>(pMsg)->level == kHighAlert
	 && static_cast<sAIAlertnessMsg*>(pMsg)->oldLevel < kHighAlert
	 && m_iMarkerObj == 0
	 && ParamGetBool(m_iObjId, "missingonalert"))
	{
		int iMarker = GoMissing();
		if (iMarker)
		{
			SInterface<IRelation> pWatchLink;
			SInterface<ILinkManager> pLM(g_pScriptManager);
			pWatchLink = pLM->GetRelationNamed("AIWatchObj");
			if (pWatchLink)
			{
				//pWatchLink->AddFull(m_iObjId, iMarker, reinterpret_cast<void*>(const_cast<sAIWatchPoint*>(&sm_AIWatchLinkDefault)));
				long lWatch = pWatchLink->Add(m_iObjId, iMarker);
				if (lWatch)
					pWatchLink->SetData(lWatch, reinterpret_cast<void*>(const_cast<sAIWatchPoint*>(&sm_AIWatchLinkDefault)));
			}
		}
	}
	return 0;
}

MSGHANDLER cScr_MissingPerson::OnRelax(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	cMultiParm mpMode;
	pPS->Get(mpMode, m_iObjId, "AI_Mode", NULL);
	if (static_cast<int>(mpMode) != kAIM_Dead && m_iMarkerObj != 0)
	{
		NotMissing(m_iMarkerObj);
		m_iMarkerObj = 0;
	}
	return 0;
}


/***********************
 * MissingMarker
 */
MSGHANDLER cScr_MissingMarker::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!stricmp(static_cast<sScrTimerMsg*>(pMsg)->name, "GoMissing"))
	{
		SService<IPropertySrv> pPS(g_pScriptManager);
		if (!pPS->Possessed(m_iObjId, "SuspObj"))
			pPS->Add(m_iObjId, "SuspObj");
		pPS->Set(m_iObjId, "SuspObj", "Is Suspicious", 1);

		m_hDelay = NULL;
		return 0;
	}
	return 1;
}

MSGHANDLER cScr_MissingMarker::OnGoMissing(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!m_hDelay)
	{
		int iTime = ParamGetTime(m_iObjId, "missing_delay");
		if (!iTime)
		{
			SService<IPropertySrv> pPS(g_pScriptManager);
			if (!pPS->Possessed(m_iObjId, "SuspObj"))
				pPS->Add(m_iObjId, "SuspObj");
			pPS->Set(m_iObjId, "SuspObj", "Is Suspicious", 1);
		}
		else if (iTime > 0)
			m_hDelay = g_pScriptManager->SetTimedMessage2(m_iObjId, "GoMissing", iTime, kSTM_OneShot, cMultiParm::Undef);
	}
	return 0;
}

MSGHANDLER cScr_MissingMarker::OnNotMissing(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_hDelay)
	{
		g_pScriptManager->KillTimedMessage(m_hDelay);
		m_hDelay = NULL;
	}
	SService<IPropertySrv> pPS(g_pScriptManager);
	if (pPS->Possessed(m_iObjId, "SuspObj"))
		pPS->Set(m_iObjId, "SuspObj", "Is Suspicious", 0);
	return 0;
}


/***********************
 * UndeadRadar
 */
bool cScr_UndeadRadar::SetLight(float fScale, int iTime, bool bPersist)
{
	int delta = iTime - m_iStimTime;
	if ((delta < 100) && (fScale <= m_fLastStim))
		return false;

	SService<IPropertySrv> pPS(g_pScriptManager);

	m_fLastStim = fScale;
	if (!bPersist && fScale <= m_fCurrent && pPS->Possessed(m_iObjId, "ScriptTiming"))
	{
		cMultiParm mpTiming;
		pPS->Get(mpTiming, m_iObjId, "ScriptTiming", NULL);
		if (delta <= static_cast<int>(mpTiming))
			return false;
	}
	if (pPS->Possessed(m_iObjId, "SelfLit"))
	{
		int iDynLight = static_cast<int>(float(m_iIntensity) * fScale);
		pPS->SetSimple(m_iObjId, "SelfLit", iDynLight);
	}
	m_fCurrent = fScale;
	m_iStimTime = iTime;
#if 0 //def DEBUG
	DebugPrintf("UndeadRadar: %.4f", fScale);
#endif
	if (fScale > 0)
	{
		int iTimeout = 666;
		if (pPS->Possessed(m_iObjId, "ScriptTiming"))
		{
			cMultiParm mpTiming;
			pPS->Get(mpTiming, m_iObjId, "ScriptTiming", NULL);
			iTimeout = static_cast<int>(mpTiming);
			if (iTimeout < 100)
				iTimeout = 100;
		}
		if (m_hRadarTimer)
			g_pScriptManager->KillTimedMessage(m_hRadarTimer);
		m_hRadarTimer = g_pScriptManager->SetTimedMessage2(m_iObjId, "RadarTimeout", iTimeout, kSTM_OneShot, cMultiParm::Undef);
	}
	return true;
}

MSGHANDLER cScr_UndeadRadar::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	int iStim = StrToObject("UndeadStim");
	if (iStim)
	{
		SService<IActReactSrv> pARSrv(g_pScriptManager);
		pARSrv->SubscribeToStimulus(m_iObjId, iStim);
	}
	else
		DebugString("No UndeadStim for radar.");
	int iDefIntensity = 0;
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	if (pPropSrv->Possessed(m_iObjId, "SelfLit"))
	{
		cMultiParm mpProp;
		pPropSrv->Get(mpProp, m_iObjId, "SelfLit", NULL);
		iDefIntensity = static_cast<int>(mpProp);
	}
	m_iIntensity.Init(iDefIntensity);
	m_iStimTime.Init(0);
	m_fLastStim.Init(0.0f);
	m_fCurrent = m_fLastStim;
	m_hRadarTimer.Init();

	return 1;
}

MSGHANDLER cScr_UndeadRadar::OnEndScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	int iStim = StrToObject("UndeadStim");
	if (iStim)
	{
		SService<IActReactSrv> pARSrv(g_pScriptManager);
		pARSrv->UnsubscribeToStimulus(m_iObjId, iStim);
	}
	return 1;
}

MSGHANDLER cScr_UndeadRadar::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		//SetLight(0.0, pMsg->time);
		m_fLastStim = 0.0f;
		m_fCurrent = 0.0f;
		m_iStimTime = pMsg->time;
		SService<IPropertySrv> pPS(g_pScriptManager);
		if (pPS->Possessed(m_iObjId, "SelfLit"))
		{
			pPS->SetSimple(m_iObjId, "SelfLit", 0);
		}
	}
	return 1;
}

MSGHANDLER cScr_UndeadRadar::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!stricmp(static_cast<sScrTimerMsg*>(pMsg)->name, "RadarTimeout"))
	{
		m_hRadarTimer = NULL;
		SetLight(m_fLastStim, pMsg->time, true);
		m_fLastStim = 0.0f;
		return 0;
	}
	return 1;
}

MSGHANDLER cScr_UndeadRadar::OnStimulus(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sStimMsg*>(pMsg)->intensity < 0)
	{
		SetLight(- static_cast<sStimMsg*>(pMsg)->intensity, pMsg->time);
	}
	return 0;
}


/***********************
 * TrigBreadcrumb
 */
int cScr_TrigBreadcrumb::DiscernRoute(int iPlayer)
{
	SService<IObjectSrv> pOS(g_pScriptManager);
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	cScrVec vMyPos, vPlayerPos, vRoutePos;
	pOS->Position(vMyPos, m_iObjId);
	pOS->Position(vPlayerPos, iPlayer);
	vPlayerPos -= vMyPos;
	vPlayerPos.Normalize();
	DebugPrintf("Player vector is (%.2f,%.2f,%.2f)", vPlayerPos.x,vPlayerPos.y,vPlayerPos.z);

	int iDestRoute = 0;
	float fDestDir = 8.0;
	std::set<int> history;

	{
		linkset lsRoutes;
		pLS->GetAll(lsRoutes, pLTS->LinkKindNamed("Route"), m_iObjId, 0);
		for (; lsRoutes.AnyLinksLeft(); lsRoutes.NextLink())
		{
			sLink sl = lsRoutes.Get();
			if (history.find(sl.dest) == history.end())
			{
				pOS->Position(vRoutePos, sl.dest);
				vRoutePos -= vMyPos;
				vRoutePos.Normalize();
				float fRouteDir = acos(vPlayerPos.Dot(vRoutePos));
				DebugPrintf(" found (%.2f,%.2f,%.2f) @%.4f", vRoutePos.x,vRoutePos.y,vRoutePos.z,fRouteDir);
				if (fRouteDir < fDestDir)
				{
					iDestRoute = sl.dest;
					fDestDir = fRouteDir;
				}
				history.insert(sl.dest);
			}
		}
	}
	{
		linkset lsRoutes;
		pLS->GetAll(lsRoutes, pLTS->LinkKindNamed("~Route"), m_iObjId, 0);
		for (; lsRoutes.AnyLinksLeft(); lsRoutes.NextLink())
		{
			sLink sl = lsRoutes.Get();
			if (history.find(sl.dest) == history.end())
			{
				pOS->Position(vRoutePos, sl.dest);
				vRoutePos -= vMyPos;
				vRoutePos.Normalize();
				float fRouteDir = acos(vPlayerPos.Dot(vRoutePos));
				DebugPrintf(" found (%.2f,%.2f,%.2f) @%.4f", vRoutePos.x,vRoutePos.y,vRoutePos.z,fRouteDir);
				if (fRouteDir < fDestDir)
				{
					iDestRoute = sl.dest;
					fDestDir = fRouteDir;
				}
				history.insert(sl.dest);
			}
		}
	}
	DebugPrintf("Player heading toward %d", iDestRoute);
	return iDestRoute;
}

MSGHANDLER cScr_TrigBreadcrumb::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPhysSrv> pPhysSrv(g_pScriptManager);
	pPhysSrv->SubscribeMsg(m_iObjId, kPhysEnterExit);

	m_iCurrentRoute.Init(0);
	m_hTimer.Init();

	return 1;
}

MSGHANDLER cScr_TrigBreadcrumb::OnEndScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPhysSrv> pPhysSrv(g_pScriptManager);
	pPhysSrv->UnsubscribeMsg(m_iObjId, kPhysEnterExit);
	return 1;
}

MSGHANDLER cScr_TrigBreadcrumb::OnPhysEnter(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sPhysMsg* pPhysMsg = static_cast<sPhysMsg*>(pMsg);
	if (pPhysMsg->transSubmod == 0)
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		object oCret;
		pOS->Named(oCret,"Creature");
		true_bool __p;
		if (m_iCurrentRoute != 0
		 && *pOS->InheritsFrom(__p, pPhysMsg->transObj, oCret))
		{
			DebugPrintf("Sending %d to %d", pPhysMsg->transObj, int(m_iCurrentRoute));
			SimplePost(m_iObjId, pPhysMsg->transObj, "Breadcrumb", static_cast<int>(m_iCurrentRoute));
		}
	}
	return 0;
}

MSGHANDLER cScr_TrigBreadcrumb::OnPhysExit(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sPhysMsg* pPhysMsg = static_cast<sPhysMsg*>(pMsg);
	if (pPhysMsg->transSubmod == 0 && pPhysMsg->transObj == StrToObject("Player"))
	{
		m_iCurrentRoute = DiscernRoute(pPhysMsg->transObj);
		SService<IPropertySrv> pPS(g_pScriptManager);
		int iFadeTime = 10000;
		if (pPS->Possessed(m_iObjId, "ScriptTiming"))
		{
			cMultiParm mpProp;
			pPS->Get(mpProp, m_iObjId, "ScriptTiming", NULL);
			iFadeTime = mpProp;
		}
		if (m_hTimer)
			g_pScriptManager->KillTimedMessage(m_hTimer);
		m_hTimer = g_pScriptManager->SetTimedMessage2(m_iObjId, "Recently", iFadeTime, kSTM_OneShot, cMultiParm::Undef);
	}
	return 0;
}

MSGHANDLER cScr_TrigBreadcrumb::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!stricmp(static_cast<sScrTimerMsg*>(pMsg)->name, "Recently"))
	{
		m_iCurrentRoute = 0;
		m_hTimer = NULL;
		return 0;
	}
	return 1;
}


/***********************
 * FollowBreadcrumb
 */
MSGHANDLER cScr_FollowBreadcrumb::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!stricmp(static_cast<sScrTimerMsg*>(pMsg)->name, "Relax"))
	{
		SService<IPropertySrv> pPS(g_pScriptManager);
		pPS->Set(m_iObjId, "AI_AlertCap", "Min level", kNoAlert);
		return 0;
	}
	return 1;
}

MSGHANDLER cScr_FollowBreadcrumb::OnNotify(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IAIScrSrv> pAISrv(g_pScriptManager);
	if (pAISrv->GetAlertLevel(m_iObjId) < kModerateAlert)
	{
		if (pAISrv->GetAlertLevel(m_iObjId) == kNoAlert)
		{
			pAISrv->SetMinimumAlert(m_iObjId, kLowAlert);
			g_pScriptManager->SetTimedMessage2(m_iObjId, "Relax", 5000, kSTM_OneShot, cMultiParm::Undef);
		}
		int iDestObj = static_cast<int>(pMsg->data);
		if (iDestObj)
		{
			true_bool __p;
			pAISrv->MakeGotoObjLoc(__p, m_iObjId, iDestObj, kNormalSpeed, kLowPriorityAction, "Breadcrumb");
		}
	}
	return 0;
}


/***********************
 * RainSFX
 */
MSGHANDLER cScr_RainSFX::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		int iRenderWeather = 1;
		SService<IDarkGameSrv> pDGS(g_pScriptManager);
		if (pDGS->ConfigIsDefined("render_weather"))
		{
			pDGS->ConfigGetInt("render_weather", &iRenderWeather);
		}

		SService<IPGroupSrv> pSFX(g_pScriptManager);
		pSFX->SetActive(m_iObjId, iRenderWeather);
	}
	return 1;
}


/***********************
 * TrackAttach
 */
void cScr_TrackAttach::SetupTrackingLinks(void)
{
	char szBuf[64];
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	linkset lsLinks;
	pLS->GetAll(lsLinks, pLTS->LinkKindNamed("~PhysAttach"), m_iObjId, 0);
	for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
	{
		sLink sl = lsLinks.Get();
		mxs_vector* pOffset = reinterpret_cast<mxs_vector*>(lsLinks.query->Data());
		if (pOffset)
			snprintf(szBuf, 64, "%.8f,%.8f,%.8f", pOffset->x, pOffset->y, pOffset->z);
		else
			strcpy(szBuf, "0,0,0");
		ParamSetString(sl.dest, "attachvec", szBuf);
	}
}

struct sAttachMatrix
{
	float rot[9];
	//float vel[9];
};

int cScr_TrackAttach::LinkIter(ILinkSrv* pLS, ILinkQuery* pLQ, IScript* pScript, void* pData)
{
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	sLink sl;
	pLQ->Link(&sl);

	const char* pszOffset = ParamGetString(sl.dest, "attachvec");
	if (!pszOffset)
		return 1;

	cScrVec vOrig;
	sscanf(pszOffset, "%f,%f,%f", &vOrig.x, &vOrig.y, &vOrig.z);

	sAttachMatrix* pMat = reinterpret_cast<sAttachMatrix*>(pData);
	cScrVec vOffset(
		(vOrig.x * pMat->rot[0]) + (vOrig.y * pMat->rot[1]) + (vOrig.z * pMat->rot[2]),
		(vOrig.x * pMat->rot[3]) + (vOrig.y * pMat->rot[4]) + (vOrig.z * pMat->rot[5]),
		(vOrig.x * pMat->rot[6]) + (vOrig.y * pMat->rot[7]) + (vOrig.z * pMat->rot[8])
		);
	pLTS->LinkSetData(pLQ->ID(), "Offset", vOffset);

	SService<IObjectSrv> pOS(g_pScriptManager);
	cScrVec vRot;
	pOS->Facing(vRot,sl.dest);
	pOS->Position(vOrig, sl.source);
	vOffset += vOrig;
	pOS->Teleport(sl.dest, vOffset, vRot, 0);

	return 1;
}

MSGHANDLER cScr_TrackAttach::OnBeginScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	m_bInitted.Init(0);
	try
	{
		SService<IDarkHookScriptService> pDH2(g_pScriptManager);
		pDH2->InstallPropHook(m_iObjId, kDHNotifyDefault, "PhysState", m_iObjId);

		if (!m_bInitted)
		{
			SetupTrackingLinks();
			m_bInitted = 1;
		}
	}
	catch (no_interface&)
	{
		DebugString("Unable to locate DarkHook service.");
	}

	return 1;
}

MSGHANDLER cScr_TrackAttach::OnEndScript(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	try
	{
		SService<IDarkHookScriptService> pDH2(g_pScriptManager);
		pDH2->UninstallPropHook(m_iObjId, "PhysState", m_iObjId);
	}
	catch (no_interface&)
	{
		DebugString("Unable to locate DarkHook service.");
	}

	return 1;
}

MSGHANDLER cScr_TrackAttach::OnNotify(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_bInitted
	 && static_cast<sDHNotifyMsg*>(pMsg)->typeDH == kDH_Property
	 && !stricmp(static_cast<sDHNotifyMsg*>(pMsg)->sProp.pszPropName, "PhysState"))
	{
		const mxs_vector* facing = 
			&reinterpret_cast<sPhysState*>(static_cast<sDHNotifyMsg*>(pMsg)->sProp.pvPropData)->facing;
		sAttachMatrix mat;
		mxs_vector i,j;
		i.x = sin(facing->x * M_PI_2 / 90.0);
		i.y = sin(facing->y * M_PI_2 / 90.0);
		i.z = sin(facing->z * M_PI_2 / 90.0);
		j.x = cos(facing->x * M_PI_2 / 90.0);
		j.y = cos(facing->y * M_PI_2 / 90.0);
		j.z = cos(facing->z * M_PI_2 / 90.0);
		mat.rot[0] = j.y*j.z; mat.rot[1] = (i.x*i.y*j.z)-(j.x*i.z); mat.rot[2] = (j.x*i.y*j.z)+(i.x*i.z);
		mat.rot[3] = j.y*i.z; mat.rot[4] = (i.x*i.y*i.z)+(j.x*j.z); mat.rot[5] = (j.x*i.y*i.z)-(i.x*j.z);
		mat.rot[6] = -i.y; mat.rot[7] = i.x*j.y; mat.rot[8] = j.x*j.y;

		IterateLinks("~PhysAttach", static_cast<sDHNotifyMsg*>(pMsg)->sProp.idObj, 0, LinkIter, this, reinterpret_cast<void*>(&mat));
	}
	return 0;
}


/***********************
 * ElevatorCar
 */
bool cScr_ElevatorCar::TweqControl(int iTweqFlags)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	if (!pPS->Possessed(m_iObjId, "StTweqJoints"))
		return false;
	pPS->Set(m_iObjId, "StTweqJoints", "Joint1AnimS", iTweqFlags);
	pPS->Set(m_iObjId, "StTweqJoints", "Joint2AnimS", iTweqFlags);
	pPS->Set(m_iObjId, "StTweqJoints", "AnimS", kTweqFlagOn);
	return true;
}

int cScr_ElevatorCar::CurrentWaypoint(void)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	linkset lsLinks;
	pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, 0);
	for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
	{
		if (lsLinks.Data() == NULL)
		{
			sLink sl = lsLinks.Get();
			return sl.dest;
		}
	}

	return 0;
}

int cScr_ElevatorCar::FindPath(int iDest)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);

	long lkPath = pLTS->LinkKindNamed("TPath");
	int iStart = CurrentWaypoint();
	sLink sl;
	link lPath;

	if (iStart == 0)
	{
		DebugPrintf("FindPath: No TPath link!");
		return 0;
	}
	if (iStart == iDest)
	{
		DebugPrintf("FindPath: Already at destination.");
		return 0;
	}
	pLS->GetOne(lPath, lkPath, iStart, 0);
	while (lPath)
	{
		pLTS->LinkGet(lPath, sl);
		if (sl.dest == iDest)
		{
			DebugPrintf("FindPath: Destination is forward.");
			return 1;
		}
		pLS->GetOne(lPath, lkPath, sl.dest, 0);
	}
	
	lkPath = -lkPath;
	pLS->GetOne(lPath, lkPath, iStart, 0);
	while (lPath)
	{
		pLTS->LinkGet(lPath, sl);
		if (sl.dest == iDest)
		{
			DebugPrintf("FindPath: Destination is backward.");
			return -1;
		}
		pLS->GetOne(lPath, lkPath, sl.dest, 0);
	}

	DebugPrintf("FindPath: Destination not in TPath.");
	return 0;
}

bool cScr_ElevatorCar::CanGoForward(void)
{
	int iStart = CurrentWaypoint();
	if (iStart)
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		true_bool bHasPath;
		pLS->AnyExist(bHasPath, pLTS->LinkKindNamed("TPath"), iStart, 0);
		return bHasPath;
	}
	return false;
}

bool cScr_ElevatorCar::CanGoBackward(void)
{
	/* Moving terrain uses just the forward TPath link to determine
	 * the speed to move to a waypoint. If there is no link to an object,
	 * it uses the default speed of '0'. Not good.
	 * StdElevator asks that you link your waypoints in a loop. But we 
	 * want to know where the start and end waypoints are. So for the 
	 * first waypoint, which normally wouldn't have a TPath link to it,
	 * a dummy marker has to be created and linked to the real first point.
	 * This function then has to look back two steps to determine if the 
	 * vator is at the "first" waypoint.
	 */
	int iStart = CurrentWaypoint();
	if (iStart)
	{
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		true_bool bHasPath;
		link lPath;
		pLS->GetOne(lPath, pLTS->LinkKindNamed("~TPath"), iStart, 0);
		if (lPath)
		{
			sLink sl;
			pLTS->LinkGet(lPath, sl);
			pLS->AnyExist(bHasPath, pLTS->LinkKindNamed("TPath"), 0, sl.dest);
			return bHasPath;
		}
	}
	return false;
}

bool cScr_ElevatorCar::OpenDoors(void)
{
#ifdef DEBUG
	DebugPrintf("ElevatorCar: Opening doors.");
#endif
	return TweqControl(kTweqFlagOn);
}

bool cScr_ElevatorCar::CloseDoors(void)
{
#ifdef DEBUG
	DebugPrintf("ElevatorCar: Closing doors.");
#endif
	return TweqControl(kTweqFlagOn | kTweqFlagReverse);
}

bool cScr_ElevatorCar::GoForward(void)
{
#ifdef DEBUG
	DebugPrintf("ElevatorCar: Moving forward.");
#endif
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	SService<IPropertySrv> pPS(g_pScriptManager);
	pPS->Set(m_iObjId, "MovingTerrain", "active", 0);
	link lCurrent;
	link lPath;
	{
		linkset lsLinks;
		pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, 0);
		for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
		{
			if (lsLinks.Data() == NULL)
			{
				lCurrent = lsLinks.Link();
				break;
			}
		}
	}
	if (! lCurrent)
	{
		DebugPrintf("GoForward: No SP link!");
		return false;
	}
	sLink sl;
	pLTS->LinkGet(lCurrent, sl);
	int iStart = sl.dest;
	pLS->GetOne(lPath, pLTS->LinkKindNamed("TPath"), iStart, 0);
	if (! lPath)
	{
		DebugPrintf("GoForward: No TPath link!");
		return false;
	}
	pLTS->LinkGet(lPath, sl);
	pLS->GetOne(lPath, pLTS->LinkKindNamed("TPathNext"), m_iObjId, 0);
	if (lPath)
	{
		pLS->Destroy(lPath);
	}
	pLS->Create(lPath, pLTS->LinkKindNamed("TPathNext"), m_iObjId, sl.dest);
	if (! lPath)
	{
		DebugPrintf("GoForward: No TPathNext link!");
		return false;
	}
	pPS->Set(m_iObjId, "MovingTerrain", "active", 1);
	pLS->Destroy(lCurrent);
	pLS->Create(lCurrent, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, sl.dest);
	pLS->BroadcastOnAllLinks(iStart, "TurnOff", pLTS->LinkKindNamed("ControlDevice"));
	return true;
}

bool cScr_ElevatorCar::GoBackward(void)
{
#ifdef DEBUG
	DebugPrintf("ElevatorCar: Moving backward.");
#endif
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	SService<IPropertySrv> pPS(g_pScriptManager);
	pPS->Set(m_iObjId, "MovingTerrain", "active", 0);
	link lCurrent;
	link lPath;
	{
		linkset lsLinks;
		pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, 0);
		for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
		{
			if (lsLinks.Data() == NULL)
			{
				lCurrent = lsLinks.Link();
				break;
			}
		}
	}
	if (! lCurrent)
	{
		DebugPrintf("GoBackward: No SP link!");
		return false;
	}
	sLink sl;
	pLTS->LinkGet(lCurrent, sl);
	int iStart = sl.dest;
	pLS->GetOne(lPath, pLTS->LinkKindNamed("~TPath"), iStart, 0);
	if (! lPath)
	{
		DebugPrintf("GoBackward: No TPath link!");
		return false;
	}
	pLTS->LinkGet(lPath, sl);
	pLS->GetOne(lPath, pLTS->LinkKindNamed("TPathNext"), m_iObjId, 0);
	if (lPath)
	{
		pLS->Destroy(lPath);
	}
	pLS->Create(lPath, pLTS->LinkKindNamed("TPathNext"), m_iObjId, sl.dest);
	if (! lPath)
	{
		DebugPrintf("GoBackward: No TPathNext link!");
		return false;
	}
	pPS->Set(m_iObjId, "MovingTerrain", "active", 1);
	pLS->Destroy(lCurrent);
	pLS->Create(lCurrent, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, sl.dest);
	pLS->BroadcastOnAllLinks(iStart, "TurnOff", pLTS->LinkKindNamed("ControlDevice"));
	return true;
}

void cScr_ElevatorCar::Halt(void)
{
#ifdef DEBUG
	DebugPrintf("ElevatorCar: Stopping.");
#endif
	m_iPathDir = 0;
	SService<IPropertySrv> pPS(g_pScriptManager);
	pPS->Set(m_iObjId, "MovingTerrain", "active", 0);
	SimplePost(m_iObjId, m_iObjId, "Stopping");
	OpenDoors();
}

bool cScr_ElevatorCar::IsMoving(void)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	cMultiParm mpProp = 0;
	pPS->Get(mpProp, m_iObjId, "MovingTerrain", "active");
	return static_cast<bool>(mpProp);
}

MSGHANDLER cScr_ElevatorCar::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		m_iPathDir = 0;
		m_iPathDest = 0;

		int iStart = 0;
		SService<ILinkSrv> pLS(g_pScriptManager);
		SService<ILinkToolsSrv> pLTS(g_pScriptManager);
		link lPath;
		pLS->GetOne(lPath, pLTS->LinkKindNamed("TPathInit"), m_iObjId, 0);
		if (lPath)
		{
			sLink sl;
			pLTS->LinkGet(lPath, sl);
			iStart = sl.dest;
			
			linkset lsLinks;
			pLS->GetAll(lsLinks, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, 0);
			for (; lsLinks.AnyLinksLeft(); lsLinks.NextLink())
			{
				if (lsLinks.Data() == NULL)
				{
					pLS->Destroy(lsLinks.Link());
				}
			}

			link lCurrent;
			pLS->Create(lCurrent, pLTS->LinkKindNamed("ScriptParams"), m_iObjId, iStart);
			
			link lPath;
			int iPath = 0;
			long flavor = pLTS->LinkKindNamed("~TPath");
			long cd = pLTS->LinkKindNamed("ControlDevice");
			pLS->GetOne(lPath, flavor, iStart, 0);
			while (lPath)
			{
				pLTS->LinkGet(lPath, sl);
				iPath = sl.dest;
				pLS->GetOne(lPath, flavor, iPath, 0);
			}
			flavor = -flavor;
			pLS->GetOne(lPath, flavor, iPath, 0);
			while (lPath)
			{
				pLTS->LinkGet(lPath, sl);
				iPath = sl.dest;
				pLS->BroadcastOnAllLinks(iPath, "TurnOff", cd);
				pLS->GetOne(lPath, flavor, iPath, 0);
			}
		}

		g_pScriptManager->SetTimedMessage2(m_iObjId, "TPathInit", 100, kSTM_OneShot, cMultiParm::Undef);
	}
	return 1;
}

MSGHANDLER cScr_ElevatorCar::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!stricmp(static_cast<sScrTimerMsg*>(pMsg)->name, "TPathInit"))
	{
		int iStart = CurrentWaypoint();
		if (!IsMoving())
		{
			if (iStart)
				CDSend("TurnOn", iStart);
			OpenDoors();
			SimplePost(m_iObjId, m_iObjId, "Stopping");
		}
		else
		{
			CloseDoors();
			SimplePost(m_iObjId, m_iObjId, "Starting");
		}
		return 0;
	}
	if (!stricmp(static_cast<sScrTimerMsg*>(pMsg)->name, "TPathNext"))
	{
		if (m_iPathDir < 0)
		{
			if (!GoBackward())
				Halt();
		}
		else if (m_iPathDir > 0)
		{
			if (!GoForward())
				Halt();
		}
		else
			Halt();
		return 0;
	}
	return 1;
}

MSGHANDLER cScr_ElevatorCar::OnTweqComplete(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (m_iPathDir != 0)
	{
		if (m_iPathDir < 0)
		{
			if (GoBackward())
				SimplePost(m_iObjId, m_iObjId, "Starting");
			else
				Halt();
		}
		else
		{
			if (GoForward())
				SimplePost(m_iObjId, m_iObjId, "Starting");
			else
				Halt();
		}
		if (m_iPathDest == 0)
			m_iPathDir = 0;
	}
	return 0;
}

MSGHANDLER cScr_ElevatorCar::OnWaypoint(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	pPS->Set(m_iObjId, "MovingTerrain", "active", 0);
	CDSend("TurnOn", static_cast<sMovingTerrainMsg*>(pMsg)->waypoint);
	if (m_iPathDir != 0
	 && m_iPathDest != static_cast<sMovingTerrainMsg*>(pMsg)->waypoint)
	{
		g_pScriptManager->SetTimedMessage2(m_iObjId, "TPathNext", 100, kSTM_OneShot, cMultiParm::Undef);
	}
	else
		Halt();

	return 0;
}

MSGHANDLER cScr_ElevatorCar::OnStarting(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	pLS->BroadcastOnAllLinks(m_iObjId, "TurnOn", pLTS->LinkKindNamed("~PhysAttach"));
	return 0;
}

MSGHANDLER cScr_ElevatorCar::OnStopping(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	SService<ILinkSrv> pLS(g_pScriptManager);
	SService<ILinkToolsSrv> pLTS(g_pScriptManager);
	pLS->BroadcastOnAllLinks(m_iObjId, "TurnOff", pLTS->LinkKindNamed("~PhysAttach"));
	return 0;
}

MSGHANDLER cScr_ElevatorCar::OnTurnOn(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!IsMoving() && CanGoForward())
	{
		m_iPathDest = 0;
		m_iPathDir = 1;
		if (!CloseDoors())
		{
			m_iPathDir = 0;
			if (GoForward())
				SimplePost(m_iObjId, m_iObjId, "Starting");
			else
				Halt();
		}
	}
	return 0;
}

MSGHANDLER cScr_ElevatorCar::OnTurnOff(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!IsMoving() && CanGoBackward())
	{
		m_iPathDest = 0;
		m_iPathDir = -1;
		if (!CloseDoors())
		{
			m_iPathDir = 0;
			if (GoBackward())
				SimplePost(m_iObjId, m_iObjId, "Starting");
			else
				Halt();
		}
	}
	return 0;
}

MSGHANDLER cScr_ElevatorCar::OnCall(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (!IsMoving())
	{
		if (0 != (m_iPathDir = FindPath(pMsg->from)))
		{
			m_iPathDest = pMsg->from;
			if (!CloseDoors())
			{
				if (GoBackward())
					SimplePost(m_iObjId, m_iObjId, "Starting");
				else
					Halt();
			}
		}
	}
	return 0;
}


/***********************
 * TrigElevator
 */
MSGHANDLER cScr_ElevatorTrigger::OnStarting(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	CDSend("TurnOn", m_iObjId);
	return 0;
}

MSGHANDLER cScr_ElevatorTrigger::OnStopping(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	CDSend("TurnOff", m_iObjId);
	return 0;
}


/***********************
 * TrapPhysical
 */
long cScr_SetRefs::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	if (!pPS->Possessed(m_iObjId, "HasRefs"))
		pPS->Add(m_iObjId, "HasRefs");
	pPS->SetSimple(m_iObjId, "HasRefs", bTurnOn);
	if (pPS->Possessed(m_iObjId, "PhysAttr"))
	{
		pPS->Set(m_iObjId, "PhysAttr", "Is Edge Trigger", !bTurnOn);
		pPS->Set(m_iObjId, "CollisionType", NULL, 
			(bTurnOn) ? kCollisionTypeBounce : kCollisionTypeNoResult );
	}
	
	return 0;
}


#if 0
/***********************
 * TrigActiveFlicker
 */
long cScr_TweqActive::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IPropertySrv> pPropSrv(g_pScriptManager);
	cMultiParm mpFlags, mpTimer;
	pPropSrv->Get(mpFlags, cScript::m_iObjId, "StTweqBlink", "AnimS");
	bool bActive = (static_cast<int>(mpFlags) & kTweqFlagOn) != 0;
	if (bActive != bTurnOn)
	{
		SService<IActReactSrv> pARSrv(g_pScriptManager);
		cMultiParm mpAction, mpType = kTweqTypeFlicker;
		if (bTurnOn)
		{
			pPropSrv->Get(mpTimer, cScript::m_iObjId, "CfgTweqBlink", "Rate");
			m_iActive = 1;
			g_pScriptManager->SetTimedMessage2(cScript::m_iObjId, "ActiveCheck", static_cast<int>(mpTimer) + 300, kSTM_OneShot, cMultiParm::Undef);
			if (!cScrTrigger::IsLocked(pMsg))
				DoTurnOn(0);
			mpAction = kTweqDoActivate;
		}
		else
		{
			m_iActive = -1;
			if (!cScrTrigger::IsLocked(pMsg))
				DoTurnOff(0);
			mpAction = kTweqDoHalt;
		}
		pARSrv->React(pARSrv->GetReactionNamed("tweq_control"), 1.0, cScript::m_iObjId, 0,
				  mpType, mpAction, cMultiParm::Undef, cMultiParm::Undef, cMultiParm::Undef, cMultiParm::Undef, cMultiParm::Undef, cMultiParm::Undef);
	}	
	return 0;
}

MSGHANDLER cScr_TweqActive::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		g_pScriptManager->SetTimedMessage2(cScript::m_iObjId, "ActiveInit", 300, kSTM_OneShot, cMultiParm::Undef);
	}
	return 1;
}

MSGHANDLER cScr_TweqActive::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	if (!stricmp(static_cast<const char*>(pTimerMsg->name), "ActiveInit"))
	{
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		cMultiParm mpFlags, mpTimer;
		pPropSrv->Get(mpFlags, cScript::m_iObjId, "StTweqBlink", "AnimS");
		if (static_cast<int>(mpFlags) & kTweqFlagOn)
		{
			pPropSrv->Get(mpTimer, cScript::m_iObjId, "CfgTweqBlink", "Rate");
			m_iActive = 1;
			g_pScriptManager->SetTimedMessage2(cScript::m_iObjId, "ActiveCheck", static_cast<int>(mpTimer) + 300, kSTM_OneShot, cMultiParm::Undef);
			if (!cScrTrigger::IsLocked(pMsg))
				DoTurnOn(0);
		}
		else
		{
			m_iActive = -1;
			if (!cScrTrigger::IsLocked(pMsg))
				DoTurnOff(0);
		}
		return 0;
	}
	else if (!stricmp(static_cast<const char*>(pTimerMsg->name), "ActiveCheck"))
	{
		if (--m_iActive == 0)
		{
			if (!cScrTrigger::IsLocked(pMsg))
				DoTurnOff(0);
		}
		return 0;
	}
	return 1;
}

MSGHANDLER cScr_TweqActive::OnTweqComplete(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sTweqMsg* pTweqMsg = static_cast<sTweqMsg*>(pMsg);
	if (pTweqMsg->Type != kTweqTypeFlicker)
		return 1;
	if (pTweqMsg->Op == kTweqOpHaltTweq)
	{
		if (m_iActive > 0)
		{
			if (!cScrTrigger::IsLocked(pMsg))
				DoTurnOff(0);
		}
		m_iActive = -1;
	}
	else if (pTweqMsg->Op == kTweqOpFrameEvent)
	{
		if (m_iActive == 0)
		{
			if (!cScrTrigger::IsLocked(pMsg))
				DoTurnOn(0);
		}
		++m_iActive;
		SService<IPropertySrv> pPropSrv(g_pScriptManager);
		cMultiParm mpTimer;
		pPropSrv->Get(mpTimer, cScript::m_iObjId, "CfgTweqBlink", "Rate");
		g_pScriptManager->SetTimedMessage2(cScript::m_iObjId, "ActiveCheck", static_cast<int>(mpTimer) + 300, kSTM_OneShot, cMultiParm::Undef);
	}
	return 0;
}
#endif


/***********************
 * TrigProximity
 */
long cScr_ProxTrig::OnSwitch(bool bTurnOn, sScrMsg* pMsg, sMultiParm* pReply)
{
	if (bTurnOn)
	{
		if (m_hTimer == NULL)
		{
			m_hTimer = g_pScriptManager->SetTimedMessage2(cScript::m_iObjId, "ProxPulse", 100, kSTM_OneShot, cMultiParm::Undef);
		}
	}
	else
	{
		if (m_hTimer != NULL)
		{
			g_pScriptManager->KillTimedMessage(m_hTimer);
			m_hTimer = NULL;
		}
	}
	return 0;
}

MSGHANDLER cScr_ProxTrig::OnSim(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	if (static_cast<sSimMsg*>(pMsg)->fStarting)
	{
		m_hTimer = g_pScriptManager->SetTimedMessage2(cScript::m_iObjId, "ProxInit", 200, kSTM_OneShot, cMultiParm::Undef);
	}
	return 1;
}

MSGHANDLER cScr_ProxTrig::OnTimer(sScrMsg* pMsg, sMultiParm* pReply, eScrTraceAction eTrace)
{
	sScrTimerMsg* pTimerMsg = static_cast<sScrTimerMsg*>(pMsg);
	if (!stricmp(static_cast<const char*>(pTimerMsg->name), "ProxInit"))
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		cScrVec vRef, vPlayer;
		object oPlayer;
		pOS->Named(oPlayer, "Player");
		pOS->Position(vPlayer, oPlayer);
		pOS->Position(vRef, cScript::m_iObjId);
		float fDist = (vPlayer - vRef).Magnitude();
		float fLimit = ParamGetFloat(cScript::m_iObjId, "proximity");
		if (fDist <= fLimit)
			DoTurnOn(0);
		else
			DoTurnOff(0);
		m_fLastDist = fDist;
		int iPulse = static_cast<int>(fabs(fDist - fLimit) * 15.0f);
		if (iPulse < 100)
			iPulse = 100;
		else if (iPulse > 10000)
			iPulse = 10000;
		m_hTimer = g_pScriptManager->SetTimedMessage2(cScript::m_iObjId, "ProxPulse", iPulse, kSTM_OneShot, cMultiParm::Undef);
	}
	else if (!stricmp(static_cast<const char*>(pTimerMsg->name), "ProxPulse"))
	{
		SService<IObjectSrv> pOS(g_pScriptManager);
		cScrVec vRef, vPlayer;
		object oPlayer;
		pOS->Named(oPlayer, "Player");
		pOS->Position(vPlayer, oPlayer);
		pOS->Position(vRef, cScript::m_iObjId);
		float fDist = (vPlayer - vRef).Magnitude();
		float fLimit = ParamGetFloat(cScript::m_iObjId, "proximity");
		if (m_fLastDist <= fLimit && fDist > fLimit)
		{
			DoTurnOff(0);
		}
		else if (m_fLastDist > fLimit && fDist <= fLimit)
		{
			DoTurnOn(0);
		}
		m_fLastDist = fDist;
		int iPulse = static_cast<int>(fabs(fDist - fLimit) * 15.0f);
		if (iPulse < 100)
			iPulse = 100;
		else if (iPulse > 10000)
			iPulse = 10000;
		m_hTimer = g_pScriptManager->SetTimedMessage2(cScript::m_iObjId, "ProxPulse", iPulse, kSTM_OneShot, cMultiParm::Undef);
	}
	return 1;
}


/***********************
 * ScaleJointTweq
 */
long cScr_JointControl::OnScale(float fValue, sScrMsg* pMsg, sMultiParm* pReply)
{
	SService<IPropertySrv> pPS(g_pScriptManager);
	cMultiParm mpJoint;
	cAnsiStr strCfg("    rate-low-high");
	cAnsiStr strJPos("Joint ");
	int iJoint = ParamGetInt(m_iObjId, "scalejoint", 1);
	if (iJoint > 1 && iJoint < 7)
	{
		strCfg.Append('0' + iJoint);
		strJPos.Append('0' + iJoint);
	}
	else
	{
		if (iJoint != 1)
			return 1;
		strJPos.Append('1');
	}
	pPS->Get(mpJoint, m_iObjId, "CfgTweqJoints", static_cast<const char*>(strCfg));
	const mxs_vector& vJoint = *static_cast<const mxs_vector*>(mpJoint);
	cMultiParm mpJointPos = (vJoint.z - vJoint.y) * fValue + vJoint.y;
	if (!pPS->Possessed(m_iObjId, "JointPos"))
		pPS->Add(m_iObjId, "JointPos");
	pPS->Set(m_iObjId, "JointPos", static_cast<const char*>(strJPos), mpJointPos);
	return 0;
}


#endif // _DARKGAME == 2
